#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Pramod Gupta (psgupta@uw.edu)
# @Date: 2020-06-10
# @Filename: mwm_rv.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee
from astropy.coordinates import Angle

from sdssdb.peewee.sdss5db.catalogdb import (AllWise, Catalog, CatalogToTIC_v8,
                                             Gaia_DR2,
                                             SDSS_APOGEE_AllStarMerge_r13,
                                             TIC_v8, TwoMassPSC)

from target_selection import log
from target_selection.cartons import BaseCarton


# See catalog.py for the name of peewee model names corresponding
# to postgres table names:
# https://github.com/sdss/sdssdb/blob/master/python/sdssdb/peewee/sdss5db/catalogdb.py

# 2.2.1. Long Baseline (Legacy Targets)

# Shorthand name: mwm_rv_long (Not an actual target class.
# Defined as a parent sample to select from for 2.2.1.x sections below)

# Simplified Description of selection criteria:
# APOGEE-1&-2 "main survey" targets with at least 3 previous epochs
# brighter than H of 12.8 that have a Gaia parallax.

# Wiki page: Binaries and Substellar Companions

# Additional source catalogs needed: Just need sdss_apogeeallstarmerge_r13

# Additional cross-matching needed: (None)

# Return columns: apogee_id, nvisits, ra, dec, pmra, pmdec, h, baseline, fields

# cadence options for these targets (list all options,
# even though no single target will receive more than one):
# (See entries for this in 2.2.1.x below)

# Pseudo SQL (optional): SELECT apogee_id, nvisits, ra, dec,
# pmra, pmdec, h, baseline, fields FROM sdss_apogeeallstarmerge_r13
# WHERE h<12.8 AND nvisits>=3 AND dist_src == 'gaia' AND
# [targflags contains 'APOGEE_SHORT' OR 'APOGEE_INTERMEDIATE' OR
# 'APOGEE_LONG' OR '*APOGEE2_*BIN_’] AS mwm_rv_long

# Implementation:

# Non-SQL implementation:

# lead contact: Nicholas Troup
# Target selection final for v0?: No

# We associate each APOGEE target with a 2MASS source since the
# vast majority of APOGEE targets were selected from the 2MASS PSC.
# APOGEE_ID is essentially the same as
# the “designation” column of the 2MASS PSC;
# We just have to strip off the “2M” in the APOGEE_ID.
#
# For example:
# sdss5db=# select designation from  catalogdb.twomass_psc limit 2;
#    designation
# ------------------
#  12034281-5740002
#  12032366-5738380
# (2 rows)
#
# sdss5db=# select apogee_id from  catalogdb.sdss_apogeeallstarmerge_r13 limit 2;
#      apogee_id
# --------------------
#  2M14044120-1550575
#  2M14033676-1554164
# (2 rows)
#
# sdss5db=# select ltrim(apogee_id,'2M') from
#  catalogdb.sdss_apogeeallstarmerge_r13 limit 2;
#       ltrim
# ------------------
#  14044120-1550575
#  14033676-1554164
# (2 rows)
#
#
# sdss5db=# select count(1) from
#  catalogdb.sdss_apogeeallstarmerge_r13 where dist_src='gaia';
#  count
# -------
#      0
# (1 row)
#
# Hence use trim(dist_src) like below:
#
# sdss5db=# select count(1) from
#  catalogdb.sdss_apogeeallstarmerge_r13 where trim(dist_src)='gaia';
#  count
# --------
#  487508
# (1 row)
#

mwm_rv_long_condition = (SDSS_APOGEE_AllStarMerge_r13.h < 12.8,
                         SDSS_APOGEE_AllStarMerge_r13.nvisits >= 3,
                         peewee.fn.trim(SDSS_APOGEE_AllStarMerge_r13.dist_src) ==
                         'gaia',
                         (SDSS_APOGEE_AllStarMerge_r13.targflags %
                          '%APOGEE_SHORT%') |
                         (SDSS_APOGEE_AllStarMerge_r13.targflags %
                          '%APOGEE_INTERMEDIATE%') |
                         (SDSS_APOGEE_AllStarMerge_r13.targflags %
                          '%APOGEE_LONG%') |
                         (SDSS_APOGEE_AllStarMerge_r13.targflags %
                          '%APOGEE2_%BIN_%'))


class MWM_RV_Long_RM_Carton(BaseCarton):
    """2.2.1.1. Long Baseline Targets for RM Plates

    Shorthand name: mwm_rv_long_rm

    Simplified Description of selection criteria:
    Select from long-baseline targets (above) within
    the 3 BHM-RM fields observable from APO
    (sexagesimal coordinates given in Pseudo SQL below)

    Wiki page: Binaries and Substellar Companions

    Additional source catalogs needed: Select from mwm_rv_long (2.2.1)

    Additional cross-matching needed: (None)

    Return columns: apogee_id, nvisits, ra, dec, pmra, pmdec, h,
    baseline, fields (Same as 2.2.1 above)

    cadence options for these targets (list all options,
    even though no single target will receive more than one):
    mwm_rv_<nn>x8 (where <nn> = number of visits to RM fields)

    Pseudo SQL (optional): SELECT apogee_id, nvisits, ra, dec,
    pmra, pmdec, h, baseline, fields FROM mwm_rv_long
    WHERE [(ra,dec) < 3 degrees on sky from
    (14:14:49 +53:05:00) OR (10:00:00 +02:12:00) OR (02:23:30 -04:15:00)]

    Implementation:

    Non-SQL implementation:

    lead contact: Nicholas Troup

    Target selection final for v0?: No
    """
    name = 'mwm_rv_long_rm'
    category = 'science'
    cadence = None
    program = 'mwm_rv'
    mapper = 'MWM'
    priority = 2571

    # peewee Model name ---> postgres table name
    # SDSS_APOGEE_AllStarMerge_r13(CatalogdbModel)--->'sdss_apogeeallstarmerge_r13'

    def build_query(self, version_id, query_region=None):
        # WHERE [(ra,dec) < 3 degrees on sky from
        # (14:14:49 +53:05:00) OR (10:00:00 +02:12:00) OR (02:23:30 -04:15:00)]

        a_ra = Angle('14h14m49s').degree
        a_dec = Angle('53d05m00s').degree

        b_ra = Angle('10h00m00s').degree
        b_dec = Angle('02d12m00s').degree

        c_ra = Angle('02h23m30s').degree
        c_dec = Angle('-04d15m00s').degree

        # In the query for cartons in mwmw_yso.py,
        # we do not need a join with Catalog
        # because the query does not use ra and dec in the WHERE clause.
        # Here, for RV cartons, we are using ra and dec in the WHERE clause in
        # q3c_radial_query() . Hence we need a join with Catalog because
        # when using coordinates in the cartons we want to always
        # use Catalog.ra and Catalog.dec.
        # This is because those coordinates have
        # all been put in a common epoch 2015.5.

        query = (Catalog
                 .select(CatalogToTIC_v8.catalogid)
                 .join(CatalogToTIC_v8,
                       on=(Catalog.catalogid == CatalogToTIC_v8.catalogid))
                 .join(TIC_v8,
                       on=(CatalogToTIC_v8.target_id == TIC_v8.id))
                 .join(TwoMassPSC,
                       on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
                 .join(SDSS_APOGEE_AllStarMerge_r13,
                       on=(TwoMassPSC.designation ==
                           peewee.fn.ltrim(SDSS_APOGEE_AllStarMerge_r13.apogee_id,
                                           '2M')))
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True,
                        *mwm_rv_long_condition,
                        peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, a_ra, a_dec, 3) |
                        peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, b_ra, b_dec, 3) |
                        peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, c_ra, c_dec, 3)))
        # Below ra, dec and radius are in degrees
        # query_region[0] is ra of center of the region
        # query_region[1] is dec of center of the region
        # query_region[2] is radius of the region
        if query_region:
            # Here we do not need a join with Catalog since query already contains
            # a join with Catalog above.
            query = (query
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))
        return query


class MWM_RV_Long_Bplates_Carton(BaseCarton):
    """ 3.2.1.2. Long Baseline Targets for Bright Time Plates

    Shorthand name: mwm_rv_long_bplates

    Simplified Description of selection criteria:
    Select from long-baseline targets (above) observable
     from APO with H brighter than 12.2 with at least 6 APOGEE
      visits within selected APOGEE fields observable during 6-month program
       (field centers in decimal degrees given below).

    Wiki page: Binaries and Substellar Companions (Page Under Construction)

    Additional source catalogs needed: Select from mwm_rv_long (2.2.1)

    Additional cross-matching needed: (None)

    Return columns: apogee_id, nvisits, ra, dec, pmra, pmdec,
     h, baseline, fields (Same as 2.2.1 above)

    cadence options for these targets (list all options,
     even though no single target will receive more than one):
      RV-6, RV-12 (plate cadences)

    Pseudo SQL (optional): SELECT apogee_id, nvisits, ra, dec,
     pmra, pmdec, h, baseline, fields FROM mwm_rv_long
      WHERE h<12.2 AND nvisits>=6 AND
       [(ra,dec)< 3 degrees on sky
       from (11.83254, 85.251) OR (47.39417, 39.50294) OR ...
        <rest of the field centers listed below>]
    RA (Degrees) DEC(Degrees)
    11.83254    85.251
    47.39417    39.50294
    72.4084    63.603
    75.35158    22.23817
    94.8905    41.2531
    111.7472    23.6726
    114.5958    37.38599
    114.5958    21
    114.7236    17.1076
    170.3706    9.9099
    195.5556    57.06683
    200.6    -13.7
    206.4    -5.5
    207.6239    2.7267
    217.9883    8.799444
    226.0916    32.3294
    232.0711    6.8184
    234.0497    86.1432
    240.575    28.09
    248.3458    -0.5336

    Implementation:

    Non-SQL implementation:

    lead contact: Nicholas Troup

    Target selection final for v0?: No
    """
    name = 'mwm_rv_long_bplates'
    category = 'science'
    cadence = None
    program = 'mwm_rv'
    mapper = 'MWM'
    priority = 2526

    # peewee Model name ---> postgres table name
    # SDSS_APOGEE_AllStarMerge_r13(CatalogdbModel)--->'sdss_apogeeallstarmerge_r13'

    def build_query(self, version_id, query_region=None):
        ra = [0] * 21
        dec = [0] * 21
        ra[1] = 11.83254;  dec[1] = 85.251     # noqa: E702, E241
        ra[2] = 47.39417;  dec[2] = 39.50294   # noqa: E702, E241
        ra[3] = 72.4084;   dec[3] = 63.603     # noqa: E702, E241
        ra[4] = 75.35158;  dec[4] = 22.23817   # noqa: E702, E241
        ra[5] = 94.8905;   dec[5] = 41.2531    # noqa: E702, E241
        ra[6] = 111.7472;  dec[6] = 23.6726    # noqa: E702, E241
        ra[7] = 114.5958;  dec[7] = 37.38599   # noqa: E702, E241
        ra[8] = 114.5958;  dec[8] = 21         # noqa: E702, E241
        ra[9] = 114.7236;  dec[9] = 17.1076    # noqa: E702, E241
        ra[10] = 170.3706; dec[10] = 9.9099    # noqa: E702, E241
        ra[11] = 195.5556; dec[11] = 57.06683  # noqa: E702, E241
        ra[12] = 200.6;    dec[12] = -13.7     # noqa: E702, E241
        ra[13] = 206.4;    dec[13] = -5.5      # noqa: E702, E241
        ra[14] = 207.6239; dec[14] = 2.7267    # noqa: E702, E241
        ra[15] = 217.9883; dec[15] = 8.799444  # noqa: E702, E241
        ra[16] = 226.0916; dec[16] = 32.3294   # noqa: E702, E241
        ra[17] = 232.0711; dec[17] = 6.8184    # noqa: E702, E241
        ra[18] = 234.0497; dec[18] = 86.1432   # noqa: E702, E241
        ra[19] = 240.575;  dec[19] = 28.09     # noqa: E702, E241
        ra[20] = 248.3458; dec[20] = -0.5336   # noqa: E702, E241

        ra_dec_condition = (
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[1], dec[1], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[2], dec[2], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[3], dec[3], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[4], dec[4], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[5], dec[5], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[6], dec[6], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[7], dec[7], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[8], dec[8], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[9], dec[9], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[10], dec[10], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[11], dec[11], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[12], dec[12], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[13], dec[13], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[14], dec[14], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[15], dec[15], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[16], dec[16], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[17], dec[17], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[18], dec[18], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[19], dec[19], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[20], dec[20], 3))

# We use *mwm_rv_long_condition to unpack the tuple mwm_rv_long_condition.
# However, ra_dec_condition is not a tuple so it does not have a * in the front.
        query = (Catalog
                 .select(CatalogToTIC_v8.catalogid)
                 .join(CatalogToTIC_v8,
                       on=(Catalog.catalogid == CatalogToTIC_v8.catalogid))
                 .join(TIC_v8,
                       on=(CatalogToTIC_v8.target_id == TIC_v8.id))
                 .join(TwoMassPSC,
                       on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
                 .join(SDSS_APOGEE_AllStarMerge_r13,
                       on=(TwoMassPSC.designation ==
                           peewee.fn.ltrim(SDSS_APOGEE_AllStarMerge_r13.apogee_id,
                                           '2M')))
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True,
                        *mwm_rv_long_condition,
                        ra_dec_condition,
                        SDSS_APOGEE_AllStarMerge_r13.h < 12.2,
                        SDSS_APOGEE_AllStarMerge_r13.nvisits >= 6))
        # Below ra, dec and radius are in degrees
        # query_region[0] is ra of center of the region
        # query_region[1] is dec of center of the region
        # query_region[2] is radius of the region
        if query_region:
            query = (query
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))
        return query


class MWM_RV_Long_FPS_Carton(BaseCarton):
    """3.2.1.3. Long Baseline Targets for FPS

    Shorthand name: mwm_rv_long_fps

    Simplified Description of selection criteria:
     Select from long-baseline targets (above) with H brighter than 11.5

    Wiki page: Binaries and Substellar Companions (Page Under Construction)

    Additional source catalogs needed: Select from mwm_rv_long (2.2.1)

    Additional cross-matching needed: (None)

    Return columns: apogee_id, nvisits, ra, dec,
     pmra, pmdec, h, baseline, fields (Same as 2.2.1 above)

    cadence options for these targets
     (list all options, even though no single target will receive more than one):
      If H>10.8 then use mwm_rv_<nn>x2, otherwise use mwm_rv_<nn>x1,
       where <nn> = 3*ceiling((18-nvisits)/3)

    Pseudo SQL (optional): SELECT apogee_id, nvisits, ra, dec,
     pmra, pmdec, h, baseline, fields FROM mwm_rv_long WHERE h<11.5

    Implementation:

    Non-SQL implementation:

    lead contact: Nicholas Troup

    Target selection final for v0?: No
    """
    name = 'mwm_rv_long_fps'
    category = 'science'
    cadence = None
    program = 'mwm_rv'
    mapper = 'MWM'
    priority = 2500

    # peewee Model name ---> postgres table name
    # SDSS_APOGEE_AllStarMerge_r13(CatalogdbModel)--->'sdss_apogeeallstarmerge_r13'

    def build_query(self, version_id, query_region=None):

        query = (Catalog
                 .select(CatalogToTIC_v8.catalogid)
                 .join(CatalogToTIC_v8,
                       on=(Catalog.catalogid == CatalogToTIC_v8.catalogid))
                 .join(TIC_v8,
                       on=(CatalogToTIC_v8.target_id == TIC_v8.id))
                 .join(TwoMassPSC,
                       on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
                 .join(SDSS_APOGEE_AllStarMerge_r13,
                       on=(TwoMassPSC.designation ==
                           peewee.fn.ltrim(SDSS_APOGEE_AllStarMerge_r13.apogee_id,
                                           '2M')))
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True,
                        *mwm_rv_long_condition,
                        SDSS_APOGEE_AllStarMerge_r13.h < 11.5))
        # Below ra, dec and radius are in degrees
        # query_region[0] is ra of center of the region
        # query_region[1] is dec of center of the region
        # query_region[2] is radius of the region
        if query_region:
            query = (query
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))
        return query


# 2.2.2. Short Baseline (Fresh Targets)
#
# Shorthand name: mwm_rv_short (Not an actual target class.
# Defined as a parent sample to select from)
#
# Simplified Description of selection criteria:
# 2MASS Targets between H = 7 and H=12.8 with the same
# selection criteria as APOGEE-2 main survey
# with the additional requirement of having a Gaia parallax.
#
# Wiki page: Binaries and Substellar Companions (Page Under Construction)
#
# Additional source catalogs needed: 2MASS, WISE, Gaia
#
# Additional cross-matching needed:
# The 3 catalogs above should be cross-matched if not already
#
# Return columns: ra, dec, pmra, pmdec, (from Gaia); h_m (from 2MASS)
#
# cadence options for these targets
# (list all options, even though no single target will
# receive more than one):  (See entries for this in 2.2.1.x below)
#
# Pseudo SQL (optional):  SELECT catalog.catalogid,
# catalog.ra, catalog.ra, catalog.dec, catalog.pmra, catalog.pmdec,
# catalog.epoch, twomass_psc.h_m, FROM catalog
# JOIN gaia_dr2_source, twomass_psc, allWise
# WHERE h_m < 12.8 AND ((j_m-k_m) - (1.5*0.918*(h_m-w2mpro-0.05))) >= 0.5
# AND (j_msigcom <= 0.1 AND h_msigcom<=0.1 AND k_msigcom <= 0.1) AND w2_sigmpro <=0.1
# AND (ph_qual= 'AAA' OR ph_qual= 'AAB' OR ph_qual= 'ABA'
# OR ph_qual= 'BAA' OR ph_qual= 'ABB' OR ph_qual= 'BAB'
# OR ph_qual= 'BBA' OR ph_qual = 'BBB')
# AND prox >= 6 AND cc_flag='000' AND gal_contam='000'
# AND (rd_flg='111' OR rd_flg='112'  OR rd_flg='121'  OR rd_flg='211'
#   OR rd_flg='122'  OR rd_flg='212'  OR rd_flg='221'  OR rd_flg='222')
# AND ext_key=Null AS mwm_rv_short
#
# Implementation:
#
# Non-SQL implementation:
#
# lead contact: Nicholas Troup
#
# Target selection final for v0?: No

mwm_rv_short_condition = (TwoMassPSC.h_m < 12.8,
                          ((TwoMassPSC.j_m - TwoMassPSC.k_m) -
                           (1.5 * 0.918 *
                            (TwoMassPSC.h_m - AllWise.w2mpro - 0.05))) >= 0.5,
                          TwoMassPSC.j_msigcom <= 0.1,
                          TwoMassPSC.h_msigcom <= 0.1,
                          TwoMassPSC.k_msigcom <= 0.1,
                          AllWise.w2sigmpro <= 0.1,
                          ((TwoMassPSC.ph_qual == 'AAA') |
                           (TwoMassPSC.ph_qual == 'AAB') |
                           (TwoMassPSC.ph_qual == 'ABA') |
                           (TwoMassPSC.ph_qual == 'BAA') |
                           (TwoMassPSC.ph_qual == 'ABB') |
                           (TwoMassPSC.ph_qual == 'BAB') |
                           (TwoMassPSC.ph_qual == 'BBA') |
                           (TwoMassPSC.ph_qual == 'BBB')),
                          TwoMassPSC.prox >= 6,
                          TwoMassPSC.cc_flg == '000',
                          TwoMassPSC.gal_contam == 0,
                          ((TwoMassPSC.rd_flg == '111') |
                           (TwoMassPSC.rd_flg == '112') |
                           (TwoMassPSC.rd_flg == '121') |
                           (TwoMassPSC.rd_flg == '211') |
                           (TwoMassPSC.rd_flg == '122') |
                           (TwoMassPSC.rd_flg == '212') |
                           (TwoMassPSC.rd_flg == '221') |
                           (TwoMassPSC.rd_flg == '222')),
                          TwoMassPSC.ext_key >> None,
                          Gaia_DR2.parallax.is_null(False))

# Notes for above mwm_rv_short_condition:
#
# (a)Above Gaia_DR2.parallax.is_null(False) is not in the pseudo SQL
# but is in the text description as in the below line from the text:
# "with the additional requirement of having a Gaia parallax.
#
# (b) TwoMassPSC.gal_contam has a type smallint.
# Hence I compare it above with 0 instead of '000'.
# (The pseudo SQL compares gal_contam with '000')
#
# (c) There are two tables with a column called ph_qual:
#
# allwise: ph_qual       | character(4)   |
# twomass_psc: ph_qual       | character(3)     |
#
# The column twomass_psc.ph_qual has type character(3) which
# is the same as the 3 character strings in the pseudo SQL
# Hence, I have used TwoMassPSC.ph_qual above.
#
# (d) The Pseudo SQL has the below condition:
# cc_flag='000'
#
# However there is no column called cc_flag.
# There are below columns with similar names:
#
# allwise: cc_flags      | character(4)   |
# twomass_psc: cc_flg        | character(3)     |
#
# The column twomass_psc.cc_flg has type character(3) which
# is the same as the 3 character strings in the pseudo SQL
# Hence, I have used TwoMassPSC.cc_flg above.
#
#


class MWM_RV_Short_RM_Carton(BaseCarton):
    """ 2.2.2.1. Short Baseline Targets for RM Plates

    Shorthand name: mwm_rv_short_rm

    Simplified Description of selection criteria: Select from short-baseline targets
     (above) within the 3 BHM-RM fields observable from APO
     (sexagesimal coordinates given in Pseudo SQL below)

    Wiki page: Binaries and Substellar Companions (Page Under Construction)

    Additional source catalogs needed: Select from mwm_rv_short (2.2.2)

    Additional cross-matching needed: (None)

    Return columns: ra, dec, pmra, pmdec, h_m

    cadence options for these targets (list all options,
     even though no single target will receive more than one):
       mwm_rv_<nn>x8 (where <nn> = number of visits to RM fields)

    Pseudo SQL (optional): SELECT catalogid, ra, dec, pmra, pmdec,
     epoch, h_m FROM mwm_rv_short
      WHERE [(ra,dec) < 3 degrees on sky from
       (14:14:49 +53:05:00) OR (10:00:00 +02:12:00) OR (02:23:30 -04:15:00)]

    Implementation:

    Non-SQL implementation:

    lead contact: Nicholas Troup

    Target selection final for v0?: No
    """
    name = 'mwm_rv_short_rm'
    category = 'science'
    cadence = None
    program = 'mwm_rv'
    mapper = 'MWM'
    priority = 2571

    def setup_transaction(self):
        pass

    def build_query(self, version_id, query_region=None):

        # WHERE [(ra,dec) < 3 degrees on sky from
        # (14:14:49 +53:05:00) OR (10:00:00 +02:12:00) OR (02:23:30 -04:15:00)]

        a_ra = Angle('14h14m49s').degree
        a_dec = Angle('53d05m00s').degree

        b_ra = Angle('10h00m00s').degree
        b_dec = Angle('02d12m00s').degree

        c_ra = Angle('02h23m30s').degree
        c_dec = Angle('-04d15m00s').degree

        log.debug('Creating temporary table for radial query...')

        radial_query = (Catalog
                        .select(Catalog.catalogid)
                        .where(peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec,
                                                          a_ra, a_dec, 3) |
                               peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec,
                                                          b_ra, b_dec, 3) |
                               peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec,
                                                          c_ra, c_dec, 3))
                        .cte('radial_query', materialized=True))

        RadialQuery = peewee.Table('mwm_rv_short_rm_temp')

        (CatalogToTIC_v8
         .select(radial_query.c.catalogid)
         .join(radial_query,
               on=(radial_query.c.catalogid == CatalogToTIC_v8.catalogid))
         .where(CatalogToTIC_v8.version_id == version_id,
                CatalogToTIC_v8.best >> True)
         .with_cte(radial_query)
         .create_table(RadialQuery.__name__, temporary=True))

        query = (CatalogToTIC_v8
                 .select(RadialQuery.c.catalogid)
                 .join(RadialQuery,
                       on=(RadialQuery.c.catalogid == CatalogToTIC_v8.catalogid))
                 .join(TIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
                 .join(Gaia_DR2, on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
                 .join_from(TIC_v8, TwoMassPSC,
                            on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
                 .join_from(TIC_v8, AllWise,
                            on=(TIC_v8.allwise == AllWise.designation))
                 .where(*mwm_rv_short_condition))

        # Below ra, dec and radius are in degrees
        # query_region[0] is ra of center of the region
        # query_region[1] is dec of center of the region
        # query_region[2] is radius of the region
        if query_region:
            query = (query
                     .where(peewee.fn.q3c_radial_query(radial_query.c.ra,
                                                       radial_query.c.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))
        return query


class MWM_RV_Short_Bplates_Carton(BaseCarton):
    """ 3.2.2.2. Short Baseline Targets for Bright Time Plates

    Shorthand name: mwm_rv_short_bplates

    Simplified Description of selection criteria:
     Select from short-baseline targets
     (above) observable from APO with H brighter than 12.2 with
      at least 100 other targets within 3 degrees that meet the other criteria.

    Wiki page: Binaries and Substellar Companions (Page Under Construction)

    Additional source catalogs needed:
     Select from mwm_rv_short (2.2.2), but need to reference mwm_rv_long_bplates

    Additional cross-matching needed: (None)

    Return columns: ra, dec, pmra, pmdec, h_m;

    cadence options for these targets
    (list all options, even though no single target will receive more than one):
     RV-12 (plate cadence)

    Pseudo SQL (optional): SELECT catalogid, ra, dec, pmra,
     pmdec, epoch, h_m FROM mwm_rv_short
      WHERE h_m<12.2 AND [(ra,dec)< 3 degrees on sky from
       (11.83254, 85.251) OR (47.39417, 39.50294) OR ...
       <rest of the field centers listed below>]
    RA (Degrees)    DEC(Degrees)
    11.83254    85.251
    47.39417    39.50294
    72.4084    63.603
    75.35158    22.23817
    94.8905 41.2531
    111.7472    23.6726
    114.5958    37.38599
    114.5958    21
    114.7236    17.1076
    170.3706    9.9099
    195.5556    57.06683
    200.6    -13.7
    206.4    -5.5
    207.6239    2.7267
    217.9883    8.799444
    226.0916    32.3294
    232.0711    6.8184
    234.0497    86.1432
    240.575    28.09
    248.3458    -0.5336

    Implementation:

    Non-SQL implementation:

    lead contact: Nicholas Troup

    Target selection final for v0?: No
"""
    name = 'mwm_rv_short_bplates'
    category = 'science'
    cadence = None
    program = 'mwm_rv'
    mapper = 'MWM'
    priority = 2576

    def build_query(self, version_id, query_region=None):

        ra = [0] * 21
        dec = [0] * 21
        ra[1] = 11.83254;  dec[1] = 85.251     # noqa: E702, E241
        ra[2] = 47.39417;  dec[2] = 39.50294   # noqa: E702, E241
        ra[3] = 72.4084;   dec[3] = 63.603     # noqa: E702, E241
        ra[4] = 75.35158;  dec[4] = 22.23817   # noqa: E702, E241
        ra[5] = 94.8905;   dec[5] = 41.2531    # noqa: E702, E241
        ra[6] = 111.7472;  dec[6] = 23.6726    # noqa: E702, E241
        ra[7] = 114.5958;  dec[7] = 37.38599   # noqa: E702, E241
        ra[8] = 114.5958;  dec[8] = 21         # noqa: E702, E241
        ra[9] = 114.7236;  dec[9] = 17.1076    # noqa: E702, E241
        ra[10] = 170.3706; dec[10] = 9.9099    # noqa: E702, E241
        ra[11] = 195.5556; dec[11] = 57.06683  # noqa: E702, E241
        ra[12] = 200.6;    dec[12] = -13.7     # noqa: E702, E241
        ra[13] = 206.4;    dec[13] = -5.5      # noqa: E702, E241
        ra[14] = 207.6239; dec[14] = 2.7267    # noqa: E702, E241
        ra[15] = 217.9883; dec[15] = 8.799444  # noqa: E702, E241
        ra[16] = 226.0916; dec[16] = 32.3294   # noqa: E702, E241
        ra[17] = 232.0711; dec[17] = 6.8184    # noqa: E702, E241
        ra[18] = 234.0497; dec[18] = 86.1432   # noqa: E702, E241
        ra[19] = 240.575;  dec[19] = 28.09     # noqa: E702, E241
        ra[20] = 248.3458; dec[20] = -0.5336   # noqa: E702, E241

        ra_dec_condition = (
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[1], dec[1], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[2], dec[2], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[3], dec[3], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[4], dec[4], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[5], dec[5], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[6], dec[6], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[7], dec[7], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[8], dec[8], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[9], dec[9], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[10], dec[10], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[11], dec[11], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[12], dec[12], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[13], dec[13], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[14], dec[14], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[15], dec[15], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[16], dec[16], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[17], dec[17], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[18], dec[18], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[19], dec[19], 3) |
            peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec, ra[20], dec[20], 3))

# We use *mwm_rv_short_condition to unpack the tuple mwm_rv_short_condition.
# However, ra_dec_condition is not a tuple so it does not have a * in the front.
        query = (Catalog
                 .select(CatalogToTIC_v8.catalogid)
                 .join(CatalogToTIC_v8,
                       on=(Catalog.catalogid == CatalogToTIC_v8.catalogid))
                 .join(TIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
                 .join(Gaia_DR2, on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
                 .switch(TIC_v8)
                 .join(TwoMassPSC, on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
                 .switch(TIC_v8)
                 .join(AllWise, on=(TIC_v8.allwise == AllWise.designation))
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True,
                        *mwm_rv_short_condition,
                        ra_dec_condition,
                        TwoMassPSC.h_m < 12.2))
        # Below ra, dec and radius are in degrees
        # query_region[0] is ra of center of the region
        # query_region[1] is dec of center of the region
        # query_region[2] is radius of the region
        if query_region:
            query = (query
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))
        return query


class MWM_RV_Short_FPS_Carton(BaseCarton):
    """3.2.2.3. Short Baseline Targets for FPS

    Shorthand name: mwm_rv_short_fps

    Simplified Description of selection criteria:
     Select from short-baseline targets (above) with H brighter than 10.8

    Wiki page: Binaries and Substellar Companions (Page Under Construction)

    Additional source catalogs needed: Select from mwm_rv_short (2.2.2)

    Additional cross-matching needed: (None)

    cadence options for these targets (list all options,
     even though no single target will receive more than one): mwm_rv_18x1

    Pseudo SQL (optional): SELECT catalogid, ra, dec,
     pmra, pmdec, epoch, h_m FROM mwm_rv_short WHERE h_m<10.8

    Implementation:

    Non-SQL implementation:

    lead contact: Nicholas Troup

    Target selection final for v0?: No
    """
    name = 'mwm_rv_short_fps'
    category = 'science'
    cadence = None
    program = 'mwm_rv'
    mapper = 'MWM'
    priority = 2510

    def build_query(self, version_id, query_region=None):

        query = (Catalog
                 .select(CatalogToTIC_v8.catalogid)
                 .join(CatalogToTIC_v8,
                       on=(Catalog.catalogid == CatalogToTIC_v8.catalogid))
                 .join(TIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
                 .join(Gaia_DR2, on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
                 .switch(TIC_v8)
                 .join(TwoMassPSC, on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
                 .switch(TIC_v8)
                 .join(AllWise, on=(TIC_v8.allwise == AllWise.designation))
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True,
                        *mwm_rv_short_condition,
                        TwoMassPSC.h_m < 10.8))
        # Below ra, dec and radius are in degrees
        # query_region[0] is ra of center of the region
        # query_region[1] is dec of center of the region
        # query_region[2] is radius of the region
        if query_region:
            query = (query
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))
        return query



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Tom Dwelly
# @Date: 2020-03-03
# @Filename: bhm_spiders_agn.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)
# derived from guide.py

# flake8: noqa
# isort: skip_file

import peewee
from peewee import JOIN
from peewee import fn
import sdssdb


'''
# for testing do this

import peewee
import sdssdb
from sdssdb.peewee.sdss5db.catalogdb import database
database.set_profile('tunnel_operations')
from target_selection.cartons.bhm_spiders_agn import BhmSpidersAgnEfedsCarton
b = BhmSpidersAgnEfedsCarton(targeting_plan='0.1.0-beta.1')
q = b.build_query(version_id=13)
for r in q[:5]: print(r))

'''



from sdssdb.peewee.sdss5db.catalogdb import (Catalog,
                                             CatalogToLegacy_Survey_DR8,
                                             Legacy_Survey_DR8,
                                             BHM_Spiders_AGN_Superset,
                                             BHM_eFEDS_Veto,
#TODO                                             CatalogToSDSS_DR16_SpecObj,
                                             SDSS_DR16_SpecObj,
                                             )


#waiting_for_psdr2# ,  PanStarrsDr2)
#waiting_for_psdr2# ,  CatalogToPanStarrsDr2)

from target_selection.cartons.base import BaseCarton
from target_selection.mag_flux import psfmag_minus_fiber2mag, AB2nMgy


## This provides the following BHM cartons:
#    bhm_spiders_agn-efeds
# and will eventually provide:
#    bhm_spiders_agn-wide
#    bhm_spiders_agn-deep
# maybe:
#    erosita-pointlike-bright-boss





class BhmSpidersAgnEfedsCarton(BaseCarton):

    '''
    SELECT * from bhm_spiders_agn_superset AS x
    INNER JOIN legacy_survey_dr8 AS ls ON  x.ls_id = ls.ls_id
    LEFT JOIN bhm_efeds_veto AS v ON q3c_join(x.opt_ra,x.opt_dec,v.plug_ra,v.plug_dec,1.0)
    WHERE x.ero_version = "efeds_c940_V2T"
    AND WHERE x.ero_det_like > X.X
    AND WHERE (ls.fiberflux_r > D.D OR
               ls.fiberflux_z > E.E)                           # faint limits
    AND WHERE ls.fibertotflux_r < CCC.C                        # bright limit - use total flux
    AND WHERE (v.plug_ra = NULL OR
               v.sn_median_all < 1.x OR
               v.zwarning > 0 OR
               v.z_err > 2e-3 OR
               v.z_err <= 0.0)
    '''

    category = 'science'
    mapper = 'BHM'
    program = 'bhm_spiders'
    tile = False


    name = 'bhm_spiders_agn-efeds'
    cadence = 'bhm_spiders_1x8'



    def build_query(self, version_id, query_region=None):

        c = Catalog.alias()
        x = BHM_Spiders_AGN_Superset.alias()
        ls = Legacy_Survey_DR8.alias()
        c2ls = CatalogToLegacy_Survey_DR8.alias()
        v = BHM_eFEDS_Veto.alias()
#TODO        c2s = CatalogToSDSS_DR16_SpecObj.alias()
        s = SDSS_DR16_SpecObj.alias()

        fiberflux_r_max = AB2nMgy(self.parameters['fibermag_r_min'])
        fiberflux_r_min = AB2nMgy(self.parameters['fibermag_r_max'])
        fiberflux_z_min = AB2nMgy(self.parameters['fibermag_z_max'])

        flux30 = AB2nMgy(30.00)

        value = peewee.Value(self.parameters.get('value', 1.0)).cast('float').alias('value')
        match_radius_spectro = self.parameters['spec_join_radius']/3600.0

        p_f = self.parameters['priority_floor']
        priority = peewee.Case(None,
                               (
                                   ((x.ero_det_like < self.parameters['det_like_for_priority']) & (s.specobjid.is_null(True)), p_f+6),
                                   ((x.ero_det_like < self.parameters['det_like_for_priority']) & (s.specobjid.is_null(False)), p_f+7),
                                   ((x.xmatch_flags == 1 ) & (s.specobjid.is_null(True)), p_f+0),
                                   ((x.xmatch_flags == 0 ) & (s.specobjid.is_null(True)), p_f+1),
                                   ((x.xmatch_flags > 1  ) & (s.specobjid.is_null(True)), p_f+2),
                                   ((x.xmatch_flags == 1 ) & (s.specobjid.is_null(False)), p_f+3),
                                   ((x.xmatch_flags == 0 ) & (s.specobjid.is_null(False)), p_f+4),
                                   ((x.xmatch_flags > 1  ) & (s.specobjid.is_null(False)), p_f+5),
                               ),
                               p_f+9) ## should never get here


        # Notes on convertion from ls_fibermag to sdss_fiber2mag:
        # https://wiki.mpe.mpg.de/eRosita/EroAGN_eFEDS/SDSSIVSpecialPlates#Estimating_SDSS_fiber2mag_.2A_from_legacysurvey_photometry

        # Notes on converting from sdss_fiber2mag to sdss_psfmag
        # https://wiki.sdss.org/display/OPS/Contents+of+targetdb.magnitude#Contentsoftargetdb.magnitude-WhatmagnitudestoputintotheplPlugMapfilesforBOSSplatetargets?

        # A flux ratio of 0.6 (roughly what is seen in all three ls bands) is a magnitude difference of
        # fiber2mag(SDSS)-fibermag(LS) = 0.55mags
        flux_ratio = {'g' : 0.60, 'r' : 0.60, 'i' : 0.60, 'z' : 0.60 }
        # Then, also add the correction from sdss_fiber2mag to sdss_psfmag: psfmag_minus_fiber2mag('filter')

        # legacysurvey fibermags - derived from fiberfluxes - with limits to avoid divide by zero errors
        magnitude_g = (psfmag_minus_fiber2mag('g') +
                       22.5-2.5*fn.log10(fn.greatest(flux30,ls.fiberflux_g*flux_ratio['g']))).cast('float')
        magnitude_r = (psfmag_minus_fiber2mag('r') +
                       22.5-2.5*fn.log10(fn.greatest(flux30,ls.fiberflux_r*flux_ratio['r']))).cast('float')
        magnitude_z = (psfmag_minus_fiber2mag('z') +
                       22.5-2.5*fn.log10(fn.greatest(flux30,ls.fiberflux_z*flux_ratio['z']))).cast('float')
        # the simplest possible interpolation
        # TODO - we can do this better
        magnitude_i = (psfmag_minus_fiber2mag('i') +
                       22.5-2.5*fn.log10(fn.greatest(flux30,
                                                     0.5*(ls.fiberflux_r+
                                                          ls.fiberflux_z)*flux_ratio['i']))).cast('float')


        query = (
            c
            .select(c.catalogid,
                    priority.alias('priority'),
                    value,
                    magnitude_g.alias("g"),
                    magnitude_r.alias("r"),
                    magnitude_i.alias("i"),
                    magnitude_z.alias("z"),
            )
            .join(c2ls)
            .join(ls)
            .join(x)
            .join(v, JOIN.LEFT_OUTER,
                  on=(fn.q3c_join(c.ra,c.dec,
                                 v.plug_ra,v.plug_dec,
                                 match_radius_spectro) &
                      (v.sn_median_all >= self.parameters['spec_sn_thresh']) &
                      (v.zwarning == 0) &
                      (v.z_err <= self.parameters['spec_z_err_thresh']) &
                      (v.z_err > 0.0)
                      )
                  )
            .join(s, JOIN.LEFT_OUTER,
                  on=(fn.q3c_join(c.ra,c.dec,
                                  s.ra,s.dec,
                                  match_radius_spectro) &
                      (s.snmedian >= self.parameters['spec_sn_thresh']) &
                      (s.zwarning == 0) &
                      (s.zerr <= self.parameters['spec_z_err_thresh']) &
                      (s.zerr > 0.0) &
                      (s.scienceprimary > 0)
                  )
            )
            .where(c.version_id == version_id,
                   c2ls.version_id == version_id,
                   c2ls.best == True)
            .where(
                (x.ero_version == self.parameters['ero_version'] ),
                (v.pk.is_null()),
                (ls.fibertotflux_r < fiberflux_r_max),
                (
                    (ls.fiberflux_r >= fiberflux_r_min) |
                    (ls.fiberflux_z >= fiberflux_z_min)
                ),
                (x.ero_det_like > self.parameters['det_like_min']),
                (
                    (
                        (x.xmatch_method == 'XPS-ML/NWAY') &
                        (x.xmatch_metric >= self.parameters['p_any_min'])
                    ) |
                    (
                        (x.xmatch_method == 'XPS-LR') &
                        (x.xmatch_metric >= self.parameters['lr_min'])
                    )
                )
            )
            .distinct([ls.ls_id])   # avoid duplicates - trust the ls_id
        )

        return query










#### some useful snippets:
'''
# example to get the listing of fields from a PeeWee model
print(catalogdb.ErositaAGNMock._meta.fields)

x = BHM_Spiders_AGN_Superset.alias()
ls = Legacy_Survey_DR8.alias()
#q14 = SDSS_DR14_QSO.alias()
c = Catalog.alias()


for f in c._meta.fields:
    print (f)
for f in x._meta.fields:
    print (f)
for f in ls._meta.fields:
    print (f)
#for f in q14._meta.fields:
#    print (f)

'''
####




#deferred#
#deferred#
#deferred#
#deferred#
#deferred# class BhmSpidersBaseCarton(BaseCarton):
#deferred#     ''' Parent class that provides the mask selections for any SPIDERs catalogue'''
#deferred#
#deferred#     name = 'bhm_spiders'
#deferred#     category = 'science'
#deferred#     mapper = 'BHM'
#deferred#     program = 'BHM-SPIDERS'
#deferred#     tile = False
#deferred#
#deferred#     # list of skymasks - move to the config file?
#deferred#     skymasks = [
#deferred#     ]
#deferred#
#deferred#
#deferred#     def post_process(self, model, **kwargs):
#deferred#         # this is where we select on mask location
#deferred#         # get the ids and coords of all of the objects in the temp table
#deferred#         cat_id = np.array(model.catalog_id[:])
#deferred#         ra = np.array(model.ra[:])
#deferred#         dec = np.array(model.dec[:])
#deferred#
#deferred#         flags = np.ones(len(cat_id), np.bool)
#deferred#
#deferred#         for sm in self.skymasks:
#deferred#             sm.apply(lon=ra, lat=dec, flags=flags)
#deferred#
#deferred#         # not sure what to return at this point - a list of tuples?
#deferred#         result = [(i,flag) for i,flag in zip(cat_id,flags)]
#deferred#
#deferred#         return result
#deferred#
#deferred#
#deferred#
#deferred# class BhmSpidersWideBaseCarton(BhmSpidersBaseCarton):
#deferred#     ''' Parent class that provides the mask slections for any SPIDER-wide catalogue'''
#deferred#
#deferred#     name = 'bhm_spiders_wide'
#deferred#     category = 'science'
#deferred#     mapper = 'BHM'
#deferred#     program = 'BHM-SPIDERS'
#deferred#     tile = False
#deferred#
#deferred#     # list of skymasks
#deferred#     skymasks = [
#deferred#         SkyMask(filename=pkg_resources.resource_filename(
#deferred#             __name__,
#deferred#             'masks/eROSITA-DE_exgal_lsdr8_or_psdr2_proc.ply'),
#deferred#                 name="spiders_wide",
#deferred#                 masktype="mangle",
#deferred#                 sense="include",
#deferred#         ),
#deferred#         SkyMask(filename=pkg_resources.resource_filename(
#deferred#             __name__,
#deferred#             'masks/rsFields-annotated-lco-deep_proc.ply'),
#deferred#                 name="spiders_deep",
#deferred#                 masktype="mangle",
#deferred#                 sense="exclude",
#deferred#         ),
#deferred#     ]
#deferred#
#deferred#
#deferred# class BhmSpidersAgnWideLsCarton(BhmSpidersWideBaseCarton):
#deferred#
#deferred#     '''
#deferred#     spiders_agn_wide_ls:
#deferred#
#deferred#     SELECT * from bhm_spiders_agn_superset AS x
#deferred#     INNER JOIN legacy_survey_dr8 AS ls ON x.ls_id = ls.ls_id
#deferred#     WHERE x.ero_version = "version_code_TBD"
#deferred#     AND WHERE x.ero_det_like > X.X
#deferred#     AND WHERE x.xmatch_metric > 0.x
#deferred#     AND WHERE (ls.fiberflux_r > A.A OR ls.fiberflux_z > B.B)
#deferred#     AND WHERE ls.fibertotflux_r < CCC.C
#deferred#     '''
#deferred#
#deferred#     name = 'bhm_spiders_agn_wide_ls'
#deferred#     cadence = 'bhm_spiders_1x4'
#deferred#
#deferred#     def build_query(self, version_id, query_region=None):
#deferred#         c = Catalog.alias()
#deferred#         x = BHM_Spiders_AGN_Superset.alias()
#deferred#         ls = Legacy_Survey_DR8.alias()
#deferred#         c2ls = CatalogToLegacy_Survey_DR8.alias()
#deferred#
#deferred#         flux_r_max =  AB2nMgy(self.parameters['mag_r_min'])
#deferred#         flux_r_min =  AB2nMgy(self.parameters['mag_r_max'])
#deferred#         flux_z_min =  AB2nMgy(self.parameters['mag_z_max'])
#deferred#         target_value = peewee.Value(self.parameters.get('value', 1.0)).alias('value')
#deferred#
#deferred#         query = (
#deferred#             c
#deferred#             .select(c.catalogid,
#deferred#                     c.ra,
#deferred#                     c.dec,
#deferred#                     c.pmra,
#deferred#                     c.pmdec,
#deferred#                     x.target_priority.alias('priority'),
#deferred#                     ls.fiberflux_g.alias('lsfiberflux_g'),
#deferred#                     ls.fiberflux_r.alias('lsfiberflux_r'),
#deferred#                     ls.fiberflux_z.alias('lsfiberflux_z'),
#deferred#             )
#deferred#             .join(c2ls)
#deferred#             .join(ls)
#deferred#             .join(x)
#deferred#             .where(c.version_id == version_id,
#deferred#                    c2ls.version_id == version_id)
#deferred#             .where(
#deferred#                 (x.ero_version == self.parameters['ero_version'] ),
#deferred#                 (ls.fibertotflux_r < flux_r_max),
#deferred#                 ((ls.fiberflux_r   > flux_r_min) |
#deferred#                  (ls.fiberflux_z > flux_z_min) ),
#deferred#                 (x.ero_det_like > self.parameters['det_like_min']),
#deferred#                 (x.xmatch_metric > self.parameters['p_any_min']),
#deferred#             )
#deferred#         )
#deferred#
#deferred#         print(f"This query will return nrows={query.count()}  (c.f. req_ntargets={self.parameters['req_ntargets']})")
#deferred#
#deferred#         return query
#deferred#
#deferred#
#deferred#
#deferred#
#deferred#
#deferred#
#deferred#
#deferred# #waiting_for_psdr2# class BhmSpidersAgnWidePsCarton(BhmSpidersWideBaseCarton):
#deferred# #waiting_for_psdr2#
#deferred# #waiting_for_psdr2#     '''
#deferred# #waiting_for_psdr2#     spiders_agn_wide_ps:
#deferred# #waiting_for_psdr2#
#deferred# #waiting_for_psdr2#     SELECT * from bhm_spiders_agn_superset AS x
#deferred# #waiting_for_psdr2#     INNER JOIN panstarrs_dr2 AS ps ON x.ps1_dr2_objid = ls.ls_id
#deferred# #waiting_for_psdr2#     WHERE x.ero_version = "version_code_TBD"
#deferred# #waiting_for_psdr2#     AND WHERE x.ero_det_like > X.X
#deferred# #waiting_for_psdr2#     AND WHERE x.xmatch_metric > 0.x
#deferred# #waiting_for_psdr2#     AND WHERE (ls.fiberflux_r > A.A OR ls.fiberflux_z > B.B)
#deferred# #waiting_for_psdr2#     AND WHERE ls.fiberflux_r < CCC.C
#deferred# #waiting_for_psdr2#     '''
#deferred# #waiting_for_psdr2#
#deferred# #waiting_for_psdr2#     name = 'bhm_spiders_agn_wide_ls'
#deferred# #waiting_for_psdr2#     cadence = 'bhm_spiders_1x4'
#deferred# #waiting_for_psdr2#
#deferred# #waiting_for_psdr2#     def build_query(self, version_id, query_region=None):
#deferred# #waiting_for_psdr2#
#deferred# #waiting_for_psdr2#         c = Catalog.alias()
#deferred# #waiting_for_psdr2#         x = BHM_Spiders_AGN_Superset.alias()
#deferred# #waiting_for_psdr2#         ps = PanStarrsDr2.alias()
#deferred# #waiting_for_psdr2#         c2ps = CatalogToPanStarrsDr2.alias()
#deferred# #waiting_for_psdr2#
#deferred# #waiting_for_psdr2#         flux_r_max =  AB2Jy(self.parameters['r_mag_min'])
#deferred# #waiting_for_psdr2#         flux_r_min =  AB2Jy(self.parameters['r_mag_max'])
#deferred# #waiting_for_psdr2#         flux_z_min =  AB2Jy(self.parameters['z_mag_max'])
#deferred# #waiting_for_psdr2#
#deferred# #waiting_for_psdr2#         query = (
#deferred# #waiting_for_psdr2#             c
#deferred# #waiting_for_psdr2#             .select(c.catalogid,
#deferred# #waiting_for_psdr2#                     c.ra,
#deferred# #waiting_for_psdr2#                     c.dec,
#deferred# #waiting_for_psdr2#                     c.pmra,
#deferred# #waiting_for_psdr2#                     c.pmdec,
#deferred# #waiting_for_psdr2#                     x.target_priority.alias('priority'),
#deferred# #waiting_for_psdr2#                     ps.g_stk_aper_flux.alias('psaperflux_g'),
#deferred# #waiting_for_psdr2#                     ps.r_stk_aper_flux.alias('psaperflux_r'),
#deferred# #waiting_for_psdr2#                     ps.i_stk_aper_flux.alias('psaperflux_i'),
#deferred# #waiting_for_psdr2#                     ps.z_stk_aper_flux.alias('psaperflux_z'))
#deferred# #waiting_for_psdr2#             .join(ps)
#deferred# #waiting_for_psdr2#             .join(c2ps)
#deferred# #waiting_for_psdr2#             .where(
#deferred# #waiting_for_psdr2#                 (ps.r_stk_aper_flux < flux_r_max) &
#deferred# #waiting_for_psdr2#                 ((ps.r_stk_aper_flux > flux_r_min) | (ps.z_stk_aper_flux > flux_z_min) ) &
#deferred# #waiting_for_psdr2#                 (tab.ero_det_like > self.parameters['det_like_min']) &
#deferred# #waiting_for_psdr2#                 (tab.xmatch_metric > self.parameters['p_any_min'])
#deferred# #waiting_for_psdr2#             )
#deferred# #waiting_for_psdr2#         )
#deferred# #waiting_for_psdr2#
#deferred# #waiting_for_psdr2#         print(f"This query will return nrows={query.count()}  (c.f. req_ntargets={self.parameters['req_ntargets']})")
#deferred# #waiting_for_psdr2#
#deferred# #waiting_for_psdr2#         return query
#deferred#
#deferred#
#deferred#
#deferred#
#deferred#
#deferred#
#deferred#
#deferred#
#deferred#
#deferred#
#deferred#
#deferred#
#deferred#
#deferred#
#deferred#
#deferred# #
#deferred# # ##############################################################
#deferred# # ##############################################################
#deferred# # ##############################################################
#deferred# # ## MOCK TARGETS ##############################################
#deferred# # class BhmSpidersAgnWideMockCarton(BaseCarton):
#deferred# #
#deferred# #     name = 'bhm_spiders_agn_wide_mock'
#deferred# #     category = 'science'
#deferred# #     survey = 'BHM'
#deferred# #     cadence = 'bhm_spiders_1x4'
#deferred# #     tile = False
#deferred# #
#deferred# #     # list of masks - remove to config file?
#deferred# #     masks = [
#deferred# #         {"name": "wide",
#deferred# #          "type": "mangle",
#deferred# #          "polarity": "include",
#deferred# #          "filename": "eROSITA-DE_exgal_lsdr8_or_psdr2_proc.ply",
#deferred# #         },
#deferred# #         {"name": "deep",
#deferred# #          "type": "mangle",
#deferred# #          "polarity": "exclude",
#deferred# #          "filename" : "rsFields-annotated-lco-deep_proc.ply",
#deferred# #         },
#deferred# #     ]
#deferred# #
#deferred# #     def build_query(self, version_id, query_region=None):
#deferred# #         '''
#deferred# #         Pure database level query - generates a super-set of potential targets
#deferred# #         '''
#deferred# #         # get the table name from the config - maybe replace this with a list of options
#deferred# #         exec(f'tab = catalogdb.{params["catalogdb_table"]}')
#deferred# #         assert tab is not None, 'Failed to locate catalogdb_table'
#deferred# #
#deferred# #         query = (tab.select(tab.gaia_dr2_source_id.alias('catalog_id'),
#deferred# #                             tab.target_ra.alias('ra'),
#deferred# #                             tab.target_dec.alias('dec'),
#deferred# #                             tab.target_pmra.alias('pmra'),
#deferred# #                             tab.target_pmdec.alias('pmdec'),
#deferred# #                             tab.target_epoch.alias('epoch'),
#deferred# #                             tab.target_mag_r.alias('magnitude_g'),
#deferred# #                             tab.target_mag_r.alias('magnitude_r'),
#deferred# #                             tab.target_mag_r.alias('magnitude_z'))
#deferred# #                  .where((tab.target_mag_r > self.parameters['r_mag_min']) &
#deferred# #                         (tab.target_mag_r < self.parameters['r_mag_max']) &
#deferred# #                         (tab.ero_det_like_0 > self.parameters['det_like_0_min'])))
#deferred# #
#deferred# #         print(f"This query will return nrows={query.count()}  (c.f. req_ntargets={self.parameters['req_ntargets']})")
#deferred# #
#deferred# #         return query
#deferred# #
#deferred# #
#deferred# #     def post_process(self, model, **kwargs):
#deferred# #         # this is where we select on mask location
#deferred# #         # get the coords of all of the objects oin the temp table
#deferred# #         ra = np.array(model.ra[:])
#deferred# #         dec = np.array(model.dec[:])
#deferred# #
#deferred# #         return True
#deferred# #
#deferred# # ## MOCK TARGETS ##############################################
#deferred# # ##############################################################
#deferred# # ##############################################################
#deferred# # ##############################################################
#deferred#
#deferred#
#deferred#
#deferred#
#deferred# def _test_xmatch_stuff():
#deferred#     # check numbers of targets in a test patch
#deferred#
#deferred#     version_id = 11
#deferred#     search_radius_deg = 0.01
#deferred#     ra0 = 135.0
#deferred#     dec0 = 1.0
#deferred#
#deferred#     c = Catalog.alias()
#deferred# #    x = BHM_Spiders_AGN_Superset.alias()
#deferred#     ls = Legacy_Survey_DR8.alias()
#deferred#     c2ls = CatalogToLegacy_Survey_DR8.alias()
#deferred#
#deferred#
#deferred#     query = (
#deferred#         c
#deferred#         .select(c.catalogid,
#deferred#                 c.ra,
#deferred#                 c.dec,
#deferred#                 c.lead,
#deferred#                 c.version,
#deferred#                 ls.ls_id,
#deferred#                 ls.ra,
#deferred#                 ls.dec,
#deferred#         )
#deferred#         .join(c2ls)
#deferred#         .join(ls)
#deferred#         .where(c.version_id == version_id,
#deferred#                c2ls.version_id == version_id)
#deferred#         .where(
#deferred#             peewee.fn.q3c_radial_query(c.ra,c.dec,
#deferred#                                        ra0, dec0,
#deferred#                                        search_radius_deg)
#deferred#         )
#deferred#     )
#deferred#
#deferred#
#deferred#     query.select().limit(1000).count()
#deferred#


'''
Exporting from the temp table

\copy (SELECT * FROM sandbox.temp_bhm_spiders_agn_efeds)  TO '/home/tdwelly/scratch/targetdb/bhm_spiders_agn_efeds.csv' with csv header
stilts tpipe in=~/scratch/targetdb/bhm_spiders_agn_efeds.csv out=~/scratch/targetdb/bhm_spiders_agn_efeds.fits ifmt=csv ofmt=fits-basic


'''



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-06-02
# @Filename: mwm_planet.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee

from sdssdb.peewee.sdss5db.catalogdb import (TESS_TOI, Catalog,
                                             CatalogToTIC_v8, TIC_v8,
                                             TwoMassPSC)

from . import BaseCarton


class MWM_TESS_Planets_Carton(BaseCarton):
    """TESS Planets.

    Definition:
        All planet-candidate TOIs + Community TOIs (CTIOs), TESS
        Candidate-Target List (CTL) selected by a TIC Priority cut.
        All cut between 7 < H < 12.

    All the candidates are compiled in catalogdb.tess_toi. Note that this list
    contains duplicates.

    """

    name = 'mwm_planet_tess'
    program = 'mwm_planet'
    category = 'science'
    mapper = 'MWM'
    cadence = None
    priority = 2600

    def build_query(self, version_id, query_region=None):

        query = (TESS_TOI
                 .select(CatalogToTIC_v8.catalogid,
                         TESS_TOI.ticid,
                         TESS_TOI.tess_disposition,
                         TwoMassPSC.h_m,
                         TESS_TOI.target_type)
                 .join(TIC_v8)
                 .join(TwoMassPSC)
                 .join_from(TIC_v8, CatalogToTIC_v8)
                 .where(TwoMassPSC.h_m > self.parameters['h_min'],
                        TwoMassPSC.h_m < self.parameters['h_max'],
                        CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True)
                 .distinct([TESS_TOI.ticid]))

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))

        return query

    def post_process(self, model, **kwargs):
        """Set priorities based on ``target_type``.

        exo_TOI: 2600
        exo_CTOI: 2605
        2min: 2010

        """

        priorities = {'exo_TOI': 2600,
                      'exo_CTOI': 2605,
                      '2min': 2610}

        for target_type, priority in priorities.items():
            (model
             .update({model.priority: priority})
             .where(model.target_type == target_type).execute())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-06-02
# @Filename: mwm_snc.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from peewee import fn

from sdssdb.peewee.sdss5db.catalogdb import (Catalog, CatalogToTIC_v8,
                                             Gaia_DR2, TIC_v8)

from . import BaseCarton


__all__ = ['MWM_SNC_100pc', 'MWM_SNC_250pc']


class MWM_SNC_100pc(BaseCarton):
    """MWM Solar Neighbourhood Census 100pc.

    Definition:

        - Divide the sky in crowded/not crowded regions via simple geometric
          cuts in (l,b).
        - Select all stars that fall within their one-sigma uncertainty within
          100pc in the not crowded regions.
        - Select all stars that fall within their one-sigma uncertainty
          within 100pc and have an astrometric_excess_noise < 2 in the
          crowded regions.

    Pseudo SQL:

        parallax-parallax_error>10 &&
        ( (astrometric_excess_noise < 2 &&
           ( (l <= 180 && b < -0.139 * l + 25 && b > 0.139 * l - 25) ||
             (l > 180 && b > -0.139 * l + 25 && b < 0.139 * l - 25) ||
             (sqrt(pow(l - 303.2, 2) + 2 * pow(b + 44.4, 2)) < 5) ||
             (sqrt(pow(l - 280.3, 2) + 2 * pow(b + 33.0, 2)) < 8) )
        ) ||
        !( (l <= 180 && b < -0.139 * l + 25 && b > 0.139 * l - 25) ||
           (l > 180 && b > -0.139 * l + 25 && b < 0.139 * l - 25) ||
           (sqrt(pow(l - 303.2, 2) + 2 * pow(b + 44.4, 2)) < 5) ||
           (sqrt(pow(l - 280.3, 2) + 2 * pow(b + 33.0, 2)) < 8) )
        )

    """

    name = 'mwm_snc_100pc'
    category = 'science'
    program = 'mwm_snc'
    mapper = 'MWM'
    cadence = 'mwm_100pc_faint_1x1'
    priority = 1800

    def build_query(self, version_id, query_region=None):

        l = Gaia_DR2.l  # noqa
        b = Gaia_DR2.b

        # Dense regions (Galactic plane, SMC, LMC).
        gal_cut = (((l <= 180) & (b < (-0.139 * l + 25)) & (b > (0.139 * l - 25))) |  # noqa
                   ((l > 180) & (b > (-0.139 * l + 25)) & (b < (0.139 * l - 25))) |
                   (fn.sqrt(fn.pow(l - 303.2, 2) + 2 * fn.pow(b + 44.4, 2)) < 5) |
                   (fn.sqrt(fn.pow(l - 280.3, 2) + 2 * fn.pow(b + 33.0, 2)) < 8))

        query = (Gaia_DR2
                 .select(Catalog.catalogid,
                         TIC_v8.gaia_int.alias('gaia_source_id'),
                         l, b)
                 .join(TIC_v8)
                 .join(CatalogToTIC_v8)
                 .where((Gaia_DR2.parallax - Gaia_DR2.parallax_error) > 10,
                        ((Gaia_DR2.astrometric_excess_noise < 2) & gal_cut) | ~(gal_cut))
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True))

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(fn.q3c_radial_query(Catalog.ra,
                                                Catalog.dec,
                                                query_region[0],
                                                query_region[1],
                                                query_region[2])))

        return query


class MWM_SNC_250pc(BaseCarton):
    """MWM Solar Neighbourhood Census 250pc.

    Definition:

        - Divide the sky in crowded/not crowded regions via simple geometric
          cuts in (l,b)
        - Apply a general cut of absolute G-band magnitude < 6mag,
          corresponding to ~G-type stars and earlier
        - Select all stars that fall within their one-sigma uncertainty
          within 250pc in the not crowded regions
        - Select all stars that fall within their one-sigma uncertainty within
          250pc and have an astrometric_excess_noise < 2 in the crowded
          regions.

    Pseudo SQL:

        (phot_g_mean_mag + 5 * log10(parallax / 1000) + 5 < 6) &&
        parallax - parallax_error > 4 &&
        ( (astrometric_excess_noise < 2 &&
           ( (l <= 180 && b < -0.139 * l + 25 && b > 0.139 * l - 25) ||
             (l > 180 && b > -0.139 * l + 25 && b < 0.139 * l - 25) ||
             (sqrt(pow(l - 303.2, 2) + 2 * pow(b + 44.4, 2)) < 5) ||
             (sqrt(pow(l - 280.3, 2) + 2 * pow(b + 33.0, 2)) < 8))) ||
          !( (l <= 180 && b < -0.139 * l + 25 && b > 0.139 * l - 25) ||
             (l > 180 && b > -0.139 * l + 25 && b < 0.139 * l - 25) ||
             (sqrt(pow(l - 303.2, 2) + 2 * pow(b + 44.4, 2)) < 5) ||
             (sqrt(pow(l - 280.3, 2) + 2 * pow(b + 33.0, 2)) < 8)
            )
        )

    """

    name = 'mwm_snc_250pc'
    category = 'science'
    program = 'mwm_snc'
    mapper = 'MWM'
    cadence = None
    priority = 1810

    def build_query(self, version_id, query_region=None):

        l = Gaia_DR2.l  # noqa
        b = Gaia_DR2.b

        colour_cut = (Gaia_DR2.phot_g_mean_mag +
                      5 * fn.log(Gaia_DR2.parallax / 1000) + 5) < 6

        gal_cut = (((l <= 180) & (b < (-0.139 * l + 25)) & (b > (0.139 * l - 25))) |  # noqa
                   ((l > 180) & (b > (-0.139 * l + 25)) & (b < (0.139 * l - 25))) |
                   (fn.sqrt(fn.pow(l - 303.2, 2) + 2 * fn.pow(b + 44.4, 2)) < 5) |
                   (fn.sqrt(fn.pow(l - 280.3, 2) + 2 * fn.pow(b + 33.0, 2)) < 8))

        query = (Gaia_DR2
                 .select(CatalogToTIC_v8.catalogid,
                         TIC_v8.gaia_int.alias('gaia_source_id'),
                         l, b)
                 .join(TIC_v8)
                 .join(CatalogToTIC_v8)
                 .where(colour_cut,
                        (Gaia_DR2.parallax - Gaia_DR2.parallax_error) > 4,
                        ((Gaia_DR2.astrometric_excess_noise < 2) & gal_cut) | ~(gal_cut))
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True))

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(fn.q3c_radial_query(Catalog.ra,
                                                Catalog.dec,
                                                query_region[0],
                                                query_region[1],
                                                query_region[2])))

        return query



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Tom Dwelly
# @Date: 2020-03-31
# @Filename: bhm_rm.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)
# derived from guide.py

# flake8: noqa
# isort: skip_file

import peewee
import sdssdb
from peewee import JOIN

#from sdssdb.peewee.sdss5db.catalogdb import (Catalog, BHM_RM_v0, CatalogToBHM_RM_v0)
from sdssdb.peewee.sdss5db.catalogdb import (Catalog,
                                             BHM_RM_v0_2,
                                             CatalogToBHM_RM_v0,
                                             CatalogToSDSS_DR16_SpecObj,
                                             SDSS_DR16_SpecObj)

from target_selection.cartons.base import BaseCarton


# Carton descriptions here:
# https://wiki.sdss.org/display/OPS/Defining+target+selection+and+cadence+algorithms#Definingtargetselectionandcadencealgorithms-ReverberationMapping(RM)

## This module provides the following BHM cartons:
#  bhm_rm_core
#  bhm_rm_known-spec
#  bhm_rm_var
#  bhm_rm_ancillary

pmsig_min = -3.0
plxsig_min = -3.0


class BhmRmBaseCarton(BaseCarton):
    '''
    This class provides common setting and the masking routines used by all RM cartons
    '''

    name = 'bhm_rm_base'
    base_name = 'bhm_rm_base'
    category = 'science'
    mapper = 'BHM'
    program = 'bhm_rm'
    cadence = 'bhm_rm_174x8'
    tile = False
    priority = None
    alias_c = None
    alias_t = None

    def get_fieldlist(self):
        '''Read the RM field centres from the yaml'''
        fieldlist = []
        base_parameters = self.config['parameters'].get(self.base_name, None)
        if base_parameters:
            fieldlist = base_parameters['fieldlist'];
        return fieldlist


    def append_spatial_query(self, query, fieldlist):
        '''extend the peewee query using a list of field centres'''
        if fieldlist is None :
            return query
        elif len(fieldlist) == 0 :
            return query

        q = False
        for f in fieldlist:
            q = ( q | peewee.fn.q3c_radial_query(self.alias_c.ra,
                                                 self.alias_c.dec,
                                                 f['racen'],
                                                 f['deccen'],
                                                 f['radius']))
        return query.where(q)


    def build_query(self, version_id, query_region=None):
        c = Catalog.alias()
        c2t = CatalogToBHM_RM_v0.alias()
        #t = BHM_RM_v0.alias()
        t = BHM_RM_v0_2.alias()
        self.alias_c = c
        self.alias_t = t
        c2s = CatalogToSDSS_DR16_SpecObj.alias()
        s = SDSS_DR16_SpecObj.alias()

        # set the Carton priority+values here - read from yaml
        #priority = peewee.Value(int(self.parameters.get('priority', 10000))).alias('priority')

        # fold in tiers of magnitude-based priority
        priority_mag_step = 0.5
        priority_mag_bright = 17.0
        priority_mag_faint = 22.0
        priority_mag_bright_known_spec = 20.5
        priority_floor = self.parameters.get('priority', 10000)
        priority = peewee.Case(
            None,
            (
                ((t.mi <= priority_mag_bright),
                 priority_floor+0),
                (((self.name == 'bhm_rm_known_spec') &
                  ~(t.field_name.contains('SDSS-RM')) &
                  (t.mi <= priority_mag_bright_known_spec)),
                 priority_floor+0),
                ((t.mi <= priority_mag_faint),
                 (priority_floor +
                  5*(1+peewee.fn.floor((t.mi-priority_mag_bright)/priority_mag_step).cast('int')))),
                ((t.mi > priority_mag_faint),
                 priority_floor+95),
            ),
            None
        ).cast('int')


        value = peewee.Value(self.parameters.get('value', 1.0)).cast('float').alias('value')
        pmra = peewee.Value(0.0).cast('float').alias('pmra')
        pmdec = peewee.Value(0.0).cast('float').alias('pmdec')
        parallax = peewee.Value(0.0).cast('float').alias('parallax')

        ### this leads to some scattered nulls in photometry - prefer to use best estimate photom instead
        # magnitude_g = peewee.Case(None,
        #                           (
        #                             ((t.optical_survey.contains('SDSS')) & (t.psfmag_sdss[1] > 0.0), t.psfmag_sdss[1]),
        #                             ((t.optical_survey.contains('PS1')) & (t.psfmag_ps1[0] > 0.0), t.psfmag_ps1[0]),
        #                             ((t.optical_survey.contains('DES')) & (t.psfmag_des[0] > 0.0), t.psfmag_des[0]),
        #                             ((t.optical_survey.contains('NSC')) & (t.mag_nsc[0] > 0.0), t.mag_nsc[0]),
        #                           ),
        #                           None) ## should never get here
        # magnitude_r = peewee.Case(None,
        #                           (
        #                             ((t.optical_survey.contains('SDSS')) & (t.psfmag_sdss[2] > 0.0), t.psfmag_sdss[2]),
        #                             ((t.optical_survey.contains('PS1')) & (t.psfmag_ps1[1] > 0.0), t.psfmag_ps1[1]),
        #                             ((t.optical_survey.contains('DES')) & (t.psfmag_des[1] > 0.0), t.psfmag_des[1]),
        #                             ((t.optical_survey.contains('NSC')) & (t.mag_nsc[1] > 0.0), t.mag_nsc[1]),
        #                           ),
        #                           None) ## should never get here
        # magnitude_i = peewee.Case(None,
        #                           (
        #                             ((t.optical_survey.contains('SDSS')) & (t.psfmag_sdss[3] > 0.0), t.psfmag_sdss[3]),
        #                             ((t.optical_survey.contains('PS1')) & (t.psfmag_ps1[2] > 0.0), t.psfmag_ps1[2]),
        #                             ((t.optical_survey.contains('DES')) & (t.psfmag_des[2] > 0.0), t.psfmag_des[2]),
        #                             ((t.optical_survey.contains('NSC')) & (t.mag_nsc[2] > 0.0), t.mag_nsc[2]),
        #                           ),
        #                           t.mi) ## should never get here
        # magnitude_z = peewee.Case(None,
        #                           (
        #                             ((t.optical_survey.contains('SDSS')) & (t.psfmag_sdss[4] > 0.0), t.psfmag_sdss[4]),
        #                             ((t.optical_survey.contains('PS1')) & (t.psfmag_ps1[3] > 0.0), t.psfmag_ps1[3]),
        #                             ((t.optical_survey.contains('DES')) & (t.psfmag_des[3] > 0.0), t.psfmag_des[3]),
        #                             ((t.optical_survey.contains('NSC')) & (t.mag_nsc[3] > 0.0), t.mag_nsc[3]),
        #                           ),
        #                           None) ## should never get here

        # Photometric precedence: DES>PS1>SDSS(>Gaia)>NSC.

        magnitude_g = peewee.Case(None,
                                  (
                                    ((t.des == 1) & (t.psfmag_des[0] > 0.0), t.psfmag_des[0]),
                                    ((t.ps1 == 1) & (t.psfmag_ps1[0] > 0.0), t.psfmag_ps1[0]),
                                    ((t.sdss == 1) & (t.psfmag_sdss[1] > 0.0), t.psfmag_sdss[1]),
                                    ((t.nsc == 1) & (t.mag_nsc[0] > 0.0), t.mag_nsc[0]),
                                  ),
                                  None) ## should never get here
        magnitude_r = peewee.Case(None,
                                  (
                                    ((t.des == 1) & (t.psfmag_des[1] > 0.0), t.psfmag_des[1]),
                                    ((t.ps1 == 1) & (t.psfmag_ps1[1] > 0.0), t.psfmag_ps1[1]),
                                    ((t.sdss == 1) & (t.psfmag_sdss[2] > 0.0), t.psfmag_sdss[2]),
                                    ((t.nsc == 1) & (t.mag_nsc[1] > 0.0), t.mag_nsc[1]),
                                  ),
                                  None) ## should never get here
        magnitude_i = peewee.Case(None,
                                  (
                                    ((t.des == 1) & (t.psfmag_des[2] > 0.0), t.psfmag_des[2]),
                                    ((t.ps1 == 1) & (t.psfmag_ps1[2] > 0.0), t.psfmag_ps1[2]),
                                    ((t.sdss == 1) & (t.psfmag_sdss[3] > 0.0), t.psfmag_sdss[3]),
                                    ((t.nsc == 1) & (t.mag_nsc[2] > 0.0), t.mag_nsc[2]),
                                  ),
                                  t.mi) ## should never get here
        magnitude_z = peewee.Case(None,
                                  (
                                    ((t.des == 1) & (t.psfmag_des[3] > 0.0), t.psfmag_des[3]),
                                    ((t.ps1 == 1) & (t.psfmag_ps1[3] > 0.0), t.psfmag_ps1[3]),
                                    ((t.sdss == 1) & (t.psfmag_sdss[4] > 0.0), t.psfmag_sdss[4]),
                                    ((t.nsc == 1) & (t.mag_nsc[3] > 0.0), t.mag_nsc[3]),
                                  ),
                                  None) ## should never get here


        ## now gaia mags
        #magnitude_bp = peewee.Case(None,
        #                           (
        #                            ((t.gaia == 1), t.mag_gaia[1]),
        #                           ),
        #                           None)
        #magnitude_rp = peewee.Case(None,
        #                           (
        #                            ((t.gaia == 1), t.mag_gaia[2]),
        #                           ),
        #                           None)



        query = (
            c
            .select(c.catalogid,
                    priority.alias('priority'),
                    value,
                    pmra,
                    pmdec,
                    parallax,
                    magnitude_g.alias('g'),
                    magnitude_r.alias('r'),
                    magnitude_i.alias('i'),
                    magnitude_z.alias('z'),
            )
            .join(c2t)
            # The following is needed because we are using c2t for Catalog_to_BHM_RM_v0
            # rather than a native c2t for Catalog_to_BHM_RM_v0_2
            .join(t, on=(c2t.target_id == t.pk))
            .where(c.version_id == version_id,
                   c2t.version_id == version_id,
                   (
                       (c2s.version_id == version_id) |
                       (c2s.version_id.is_null())
                   ),
                   c2t.best == True)
            .where
            (
                (t.mi >= self.parameters['mag_i_min']),
                (t.mi <  self.parameters['mag_i_max']),
            )
            .switch(c)
            .join(c2s, JOIN.LEFT_OUTER)
            .join(s, JOIN.LEFT_OUTER)
            .where(
                (s.specobjid.is_null()) |    # no match in sdss_dr16_specobj
                ~(
                    (s.class_.contains('STAR')) & # Reject objects where the best spectrum for
                    (s.scienceprimary > 0)      # this object is classified as STAR
                )
            )
            .distinct([t.pk])   # avoid duplicates - trust the RM parent sample
        )
        query = self.append_spatial_query(query, self.get_fieldlist())

        return query




class BhmRmCoreCarton(BhmRmBaseCarton):
    '''
    bhm_rm_core: select all photometric QSO targets with the likelihood method (Skewt), flux-limited to 21.5 in i-band PSF mag

    SELECT * FROM bhm_rm
            WHERE skewt_qso = 1
            AND WHERE mi BETWEEN 15.0 AND 21.5
            AND WHERE pmsig < 5.0
            AND WHERE plxsig < 5.0


    also require t.skewt_qso_prior == 1 in CVZ-S
    '''

    name = 'bhm_rm_core'

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        t = self.alias_t
        query = query.where(
            (t.skewt_qso == 1) &
            (
                ~(t.field_name.contains('S-CVZ')) |
                ((t.field_name.contains('S-CVZ')) & (t.skewt_qso_prior == 1))
            ) &
            (t.pmsig <  self.parameters['pmsig_max']) &
            (t.plxsig <  self.parameters['plxsig_max']) &
            (
                #catch bad photometry - require at least gri detections in at least one system
                ((t.sdss == 1) &
                 ( t.psfmag_sdss[1] > 0.0) &
                 ( t.psfmag_sdss[2] > 0.0) &
                 ( t.psfmag_sdss[3] > 0.0)) |
                ((t.ps1 == 1) &
                 ( t.psfmag_ps1[0] > 0.0) &
                 ( t.psfmag_ps1[1] > 0.0) &
                 ( t.psfmag_ps1[2] > 0.0)) |
                ((t.des == 1) &
                 ( t.psfmag_des[0] > 0.0) &
                 ( t.psfmag_des[1] > 0.0) &
                 ( t.psfmag_des[2] > 0.0)) |
                ((t.nsc == 1) &
                 ( t.mag_nsc[0] > 0.0) &
                 ( t.mag_nsc[1] > 0.0) &
                 ( t.mag_nsc[2] > 0.0))
            ) &
#            (t.pmsig > pmsig_min) &  # this catches cases with NULL=-9
#            (t.plxsig > plxsig_min) &
            ~(t.field_name.contains('SDSS-RM')) # ignore this carton in the SDSS-RM field
        )

        return query

class BhmRmKnownSpecCarton(BhmRmBaseCarton):
    '''
    bhm_rm_known-spec:  select all spectroscopically confirmed QSOs where redshift is extragalactic

    SELECT * FROM bhm_rm
        WHERE specz > 0.005
        AND WHERE mi BETWEEN 15.0 AND 21.5   # <- exact limits TBD

    For SDSS-RM field select only QSOs that were part of the original SDSS-III/IV RM program
    Do that via a bit in the spec_bitmask field. Also restrict to mi < 21.0
    '''

    name = 'bhm_rm_known-spec'

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        t = self.alias_t
        spec_bitmask_sdss_rm_qso = 2**3
        query = query.where(
            (t.spec_q == 1 ) &
            (
                ~(t.field_name.contains('SDSS-RM')) |
                (
                    # include extra constraints on SDSS-RM targets
                    (t.spec_bitmask.bin_and(spec_bitmask_sdss_rm_qso) != 0) &
                    (t.mi < self.parameters['mag_i_max_sdss_rm'] )
                )
            ) &
            (
                ~(t.field_name.contains('COSMOS')) |
                (
                    # include extra constraints on COSMOS targets
                    (t.mi < self.parameters['mag_i_max_cosmos'] )
                )
            ) &
            (
                ~(t.field_name.contains('XMM-LSS')) |
                (
                    # include extra constraints on XMM-LSS targets
                    (t.mi < self.parameters['mag_i_max_xmm_lss'] )
                )
            ) &
            (t.specz >= self.parameters['specz_min']) &
            (t.specz <= self.parameters['specz_max'])
        )

        return query




class BhmRmVarCarton(BhmRmBaseCarton):
    '''bhm_rm_var: selected based on g-band variability > 0.05 mag and bright enough to be detected by Gaia (G<~21)

    SELECT * FROM bhm_rm
        WHERE ( (des_var_sn[0] > 5.0 AND des_var_rms[0] > 0.05)  OR
                (ps1_var_sn[0]>5.0 AND ps1_var_rms[0]>0.05))
        AND WHERE mi BETWEEN 15.0 AND 21.5    # <- exact limits TBD
        AND WHERE pmsig < x.x
        AND WHERE plxsig < x.x
        AND WHERE gaia = 1

    #debug select t.pk, t.ra, t.dec, t.mi, t.psfmag_sdss[4] as
    psfmag_i,t.pmsig,t.ps1_var_sn[1],t.ps1_var_rms[1],t.des_var_sn[1],t.des_var_rms[1]
    from bhm_rm_v0 as t where (t.gaia = 1 AND t.mi < 21.5 AND t.pmsig
    < 5.0 AND t.plxsig < 5.0 AND t.ps1_var_sn[1] > 5.0 AND
    t.ps1_var_rms[1] > 0.05 ) limit 10;

    '''

    name = 'bhm_rm_var'

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        t = self.alias_t
        query = query.where(
            (
                (
                    (t.des_var_sn[0] > self.parameters['des_var_sn_min']) &
                    (t.des_var_rms[0] > self.parameters['des_var_rms_min'])
                ) |
                (
                    (t.ps1_var_sn[0] > self.parameters['ps1_var_sn_min']) &
                    (t.ps1_var_rms[0] > self.parameters['ps1_var_rms_min'])
                )
            ) &
            (t.source_id_gaia > 0 ) &
            (t.pmsig < self.parameters['pmsig_max']) &
            (t.plxsig < self.parameters['plxsig_max']) &
            (t.pmsig > pmsig_min) &  # this catches cases with NULL=-9
            (t.plxsig > plxsig_min) &
            ~(t.field_name.contains('SDSS-RM')) # ignore this carton in the SDSS-RM field
        )
            #& (t.mg <  self.parameters['g_mag_max'])  # TBD

        return query



class BhmRmAncillaryCarton(BhmRmBaseCarton):
    '''
    bhm_rm_ancillary: from the Gaia_unWISE AGN catalog or the XDQSO catalog,
                      but requiring no proper motion/parallax detection from Gaia DR2

    SELECT * FROM bhm_rm
        WHERE photo_bitmask & (2^0+2^1) != 0
        AND WHERE mi BETWEEN 15.0 AND 21.5    ← TBD
        AND WHERE pmsig < 5.0
        AND WHERE plxsig < 5.0
    '''

    name = 'bhm_rm_ancillary'

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        t = self.alias_t

        query = query.where(
            (t.photo_bitmask.bin_and(self.parameters['photo_bitmask']) != 0 ) &
            (t.pmsig <  self.parameters['pmsig_max']) &
            (t.plxsig <  self.parameters['plxsig_max']) &
            (t.pmsig > pmsig_min) &  # this catches cases with NULL=-9
            (t.plxsig > plxsig_min) &
            ~(t.field_name.contains('SDSS-RM')) # ignore this carton in the SDSS-RM field
        )

        return query










#######################################################################
#######################################################################
#######################################################################
#######################################################################
# Notes and debug
#
# '''
# t = BHM_RM_v0.alias()
# for f in t._meta.fields:
#     print (f)
# '''
#
# sql, param = qq.sql()
# print (sql.replace("%s","{}").format(*param))
#    # some functional SQL for one field:
#    ##SELECT c.*,t.specz,t.mi FROM catalog AS C INNER JOIN catalog_to_bhm_rm_v0 AS c2t ON c.catalogid = c2t.catalogid INNER JOIN bhm_rm_v0 AS t ON c2t.target_id = t.pk WHERE (q3c_radial_query(c.ra, c.dec, 36.45, -4.6, 1.49) AND c.version_id = 13 AND t.specz > 0.005 AND mi < 21.1 AND mi > 16.0);
#    SELECT DISTINCT ON (t.pk) c.*,t.mg,t.mi,t.specz,t.spec_q,t.spec_strmask,specz_ref,q3c_dist(c.ra,c.dec,t.ra,t.dec)*3600.::FLOAT as sep FROM catalog AS C INNER JOIN catalog_to_bhm_rm_v0 AS c2t ON c.catalogid = c2t.catalogid INNER JOIN bhm_rm_v0 AS t ON c2t.target_id = t.pk WHERE (q3c_radial_query(c.ra, c.dec, 36.45, -4.6, 1.49) AND c.version_id = 13 AND t.specz > 0.005 AND mi < 21.1 AND mi > 16.0);
#    \copy (SELECT DISTINCT ON (t.pk) c.*,t.mg,t.mi,t.specz,t.spec_q,t.spec_strmask,specz_ref,q3c_dist(c.ra,c.dec,t.ra,t.dec)*3600.::FLOAT as sep FROM catalog AS C INNER JOIN catalog_to_bhm_rm_v0 AS c2t ON c.catalogid = c2t.catalogid INNER JOIN bhm_rm_v0 AS t ON c2t.target_id = t.pk WHERE (q3c_radial_query(c.ra, c.dec, 35.5, -4.6, 1.49) AND c.version_id = 13 AND t.specz > 0.005 AND mi < 22.0 AND mi > 16.0)) to ./test_faint.csv csv header;
#



'''
# for testing do domething like this

import peewee
import sdssdb
from sdssdb.peewee.sdss5db.catalogdb import database
database.set_profile('tunnel_operations')
from target_selection.cartons.bhm_rm import *
c = BhmRmCoreCarton(targeting_plan='0.1.0-beta.1')
c = BhmRmKnownSpecCarton(targeting_plan='0.1.0-beta.1')
c = BhmRmVarCarton(targeting_plan='0.1.0-beta.1')
c = BhmRmAncillaryCarton(targeting_plan='0.1.0-beta.1')
q = c.build_query(version_id=13)
q.count()
for r in q.limit(5).namedtuples():
    print(r)

'''


'''
#target_selection  --profile tunnel_operations --verbose run --include bhm_rm_known_spec --keep --overwrite '0.1.0-beta.1' --no-load

target_selection  --profile tunnel_operations --verbose run --include bhm_rm_core,bhm_rm_var,bhm_rm_ancillary,bhm_rm_known_spec --keep --overwrite '0.1.0-beta.1' --no-load

# to get a psql prompt:
#> psql -d sdss5db -U sdss_user -h localhost -p 7502

# Exporting from the temp table
# in psql terminal:

\copy (SELECT * FROM sandbox.temp_bhm_rm_core)  TO '/home/tdwelly/scratch/targetdb/bhm_rm_core.csv' with csv header
\copy (SELECT * FROM sandbox.temp_bhm_rm_known_spec)  TO '/home/tdwelly/scratch/targetdb/bhm_rm_known_spec.csv' with csv header
\copy (SELECT * FROM sandbox.temp_bhm_rm_var)  TO '/home/tdwelly/scratch/targetdb/bhm_rm_var.csv' with csv header
\copy (SELECT * FROM sandbox.temp_bhm_rm_ancillary)  TO '/home/tdwelly/scratch/targetdb/bhm_rm_ancillary.csv' with csv header

for F in bhm_rm_*.csv; do   stilts tpipe in=${F} out="${F%.*}.fits" ifmt=csv ofmt=fits-basic; done

head -n 1 bhm_rm_core.csv  > bhm_rm.csv
tail -q -n +2 bhm_rm_*.csv  >> bhm_rm.csv
stilts tpipe in=bhm_rm.csv out=bhm_rm.fits ifmt=csv ofmt=fits-basic
ftcopy "bhm_rm.fits[1][col *,bhm_rm_core(L)=priority==1002?1:0,bhm_rm_known_spec(L)=priority==1001?1:0,bhm_rm_var(L)=priority==1003?1:0,bhm_rm_ancillary(L)=priority==1004?1:0]" bhm_rm.fits clobber=yes mode=q
ftsort bhm_rm.fits bhm_rm_unique.fits catalogid method=heap unique=yes clobber=yes mode=q

gnuplot plot_rm_target_stats.plot
tar -cvzf bhm_core_cartons.tar.gz bhm_rm*.fits bhm_aqmes_*fits bhm_spiders_*fits bhm_csc_*fits

'''

#######################################################################
#######################################################################
#######################################################################



import glob
import importlib
import inspect
import os
import warnings

from ..exceptions import TargetSelectionImportWarning
from .base import BaseCarton


# Import cartons so that they can be discovered by
# calling Carton.__subclasses__().

exclusions = ['__init__.py', 'base.py']

cwd = os.getcwd()
os.chdir(os.path.dirname(os.path.realpath(__file__)))

files = [file_ for file_ in glob.glob('**/*.py', recursive=True)
         if file_ not in exclusions]

for file_ in files:

    try:

        modname = file_[0:-3].replace('/', '.')
        mod = importlib.import_module('target_selection.cartons.' + modname)
        for objname in dir(mod):
            obj = getattr(mod, objname)
            if inspect.isclass(obj) and issubclass(obj, BaseCarton):
                locals().update({objname: obj})

    except Exception as ee:

        warnings.warn(f'cannot import file {file_}: {ee}',
                      TargetSelectionImportWarning)

os.chdir(cwd)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-05-06
# @Filename: skies.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import multiprocessing
import os
import warnings
from functools import partial

import healpy
import numpy
import pandas
from astropy.coordinates import SkyCoord, match_coordinates_sky
from matplotlib import pyplot as plt
from matplotlib.patches import Ellipse

from target_selection import log, manager
from target_selection.exceptions import TargetSelectionUserWarning


def nested_regrade(pixels, nside_in, nside_out):
    r"""Returns the parent/children pixels from a given HealPix nested pixel.

    The HealPix nested mode follows a quadrilateral tree pixel scheme (see
    Figure 1 in Górski et al. 2005) When the resolution (nside) increases,
    each pixel is divided in four pixels. The numbering of such pixels follows
    a simple binary representation achieved by appending binary digits to the
    new pixels. So, for example, pixel 22 (``b10110``) becomes four new pixels
    with binary numbers ``b1011000, b1011001, b1011010, and b1011011``,
    corresponding to decimal ``91, 92, 93, and 94``.

    This function returns the child pixels from a given pixel when going from
    resolution ``nside_in`` to ``nside_out``, if ``nside_out > nside_in``, or
    the parent pixel if ``nside_out < nside_in``.

    Note that this function works only for pixels using the nested indexing and
    should not be used with the ring indexing.

    Parameters
    ----------
    pixels : int or ~numpy.ndarray
        The pixels for which we want to get the parents/children. Can be a
        single integer or an array of indices.
    nside_in : int
        The ``nside`` of the input pixels. Must be one of :math:`2^k` where
        :math:`k\in[0,1,2,\dots]`.
    nside_out : int
        The destination ``nside``.

    Returns
    -------
    output : `int` or `~numpy.ndarray`
        If ``nside_out < nside_in`` and ``pixels`` is a single value, this
        will be an integer with the parent of the input pixel. If ``pixels``
        is an array the output will be an array of the same size in which each
        element is the parent of the corresponding pixel in the input array.
        If ``nside_out > nside_in`` and ``pixels`` is an integer, the output
        will be an array with the child pixels. The size of the array will be
        :math:`2^{2l}` where ``l=k_out-k_in`` (:math:`n_{side}=2^k`).
        If the input in an array of pixels, the output will be a 2D array in
        which each row contains the child pixels of the corresponding pixel
        in the input array.

    """

    assert nside_in != nside_out, 'nside_in cannot be equal to nside_out.'

    k_in = numpy.log2(nside_in)
    k_out = numpy.log2(nside_out)

    assert k_in.is_integer() and k_out.is_integer(), \
        'nside_in or nside_out are not power of 2.'

    pixels = numpy.atleast_1d(pixels).astype(int)
    assert pixels.ndim == 1, 'dimension of input pixels is invalid.'

    # Npix = 12 * nside**2
    assert numpy.all(pixels <= 12 * nside_in**2), \
        'some pixel indices are greater than the maximum allowed for nside_in.'

    if k_in > k_out:

        degraded = numpy.right_shift(pixels, 2 * int(k_in - k_out))

        return degraded[0] if len(degraded) == 1 else degraded

    else:

        prograded = numpy.zeros((len(pixels), int(2**(2 * (k_out - k_in)))),
                                dtype=int)
        prograded += pixels[numpy.newaxis].T
        for ii in range(int(k_out - k_in))[::-1]:

            prograded = numpy.left_shift(prograded, 2)

            repeats = numpy.repeat([0, 1, 2, 3], (4 ** ii) or 1)
            tiles = numpy.tile(repeats, prograded.shape[1] // len(repeats))

            prograded += tiles

        return prograded


def _process_tile(inputs, candidate_nside=None, tile_nside=None,
                  min_separation=None, mag_column=None, is_flux=False,
                  mag_threshold=None, downsample=False,
                  downsample_nside=None, seed=None):
    """Processes a tile from catalogue data."""

    tile, targets = inputs

    if len(targets) == 0:
        return False

    # For each target we calculate the corresponding pixel in candidate_nside
    # resolution.
    cpcol = f'pix_{candidate_nside}'
    targets[cpcol] = healpy.ang2pix(candidate_nside,
                                    targets.ra, targets.dec,
                                    nest=True, lonlat=True)

    # Get all the pixels in candidate_nside resolution that correspond to this
    # tile. Create a mask with all of them that are not occupied by a target.
    all_pix = nested_regrade(tile, tile_nside, candidate_nside)[0]
    mask_not_target = ~numpy.in1d(all_pix, targets[cpcol])

    # Create a DF with the candidate pixels and calculate their RAs and Decs.
    candidates = pandas.DataFrame(all_pix[mask_not_target], columns=[cpcol])
    candidates['ra'], candidates['dec'] = healpy.pix2ang(candidate_nside,
                                                         candidates[cpcol],
                                                         nest=True,
                                                         lonlat=True)

    # If we are using the magnitudes to correct the minimum separation, we
    # first convert the column to magnitudes if it's flux. Then select all the
    # targets with magnitude < mag_threshold and for each one mask out all
    # the candidate pixels that lie within the corrected separation.
    # TODO: We use a rectangular search here because doing a
    # SkyCoord.separation for each target would be very costly but there may
    # be a better way.
    if mag_column:

        if is_flux:
            mask_flux = targets[mag_column] > 0
            targets.loc[mask_flux, mag_column] = (
                22.5 - 2.5 * numpy.log10(targets[mask_flux][mag_column]))
            targets.loc[targets[mag_column] <= 0, mag_column] = numpy.nan

        btargets = targets[targets[mag_column] < mag_threshold]
        mask = numpy.ones(len(candidates), dtype=numpy.bool)

        for _, btarget in btargets.iterrows():
            sep_corr = min_separation + numpy.sqrt((mag_threshold -
                                                    btarget[mag_column]) / 0.2)
            sep_corr /= 3600.
            sep_corr_ra = sep_corr / numpy.cos(numpy.radians(btarget.dec))
            masked_out = ((candidates.ra > (btarget.ra - sep_corr_ra)) &
                          (candidates.ra < (btarget.ra + sep_corr_ra)) &
                          (candidates.dec > (btarget.dec - sep_corr)) &
                          (candidates.dec < (btarget.dec + sep_corr)))
            mask[masked_out] = False

        candidates = candidates.loc[mask, :].copy()

    # Determine the closest neighbout for each candidate pixel.
    c1 = SkyCoord(ra=candidates.ra.to_numpy(),
                  dec=candidates.dec.to_numpy(), unit='deg')
    c2 = SkyCoord(ra=targets.ra.to_numpy(),
                  dec=targets.dec.to_numpy(), unit='deg')

    target_idx, sep, _ = match_coordinates_sky(c1, c2, nthneighbor=1)

    matched_target = targets.iloc[target_idx, :]
    sep_arcsec = sep.value * 3600.

    # Add column with the separation.
    candidates['sep_neighbour'] = sep_arcsec

    if mag_column:
        candidates['mag_neighbour'] = matched_target.loc[:, mag_column].to_numpy()

    valid_skies = candidates.loc[candidates.sep_neighbour > min_separation, :].copy()
    valid_skies[f'tile_{tile_nside}'] = tile

    # If we are not downsampling, set the index and return.
    if downsample is False or downsample is None:
        if len(valid_skies) > 0:
            valid_skies.set_index(cpcol, inplace=True)
        else:
            return False
        return valid_skies

    # If downsampling, regrade each candidate pixel value to the nside we'll
    # use for downsampling.
    valid_skies['down_pix'] = nested_regrade(valid_skies[cpcol],
                                             candidate_nside,
                                             downsample_nside)

    if isinstance(downsample, int):

        assert downsample > 1, 'downsample must be > 1'

        # Calculate how many skies per downsample pixel we should get.
        # Each next order of a pixel has four nested pixels.
        k_tile = int(numpy.log2(tile_nside))
        k_downsample = int(numpy.log2(downsample_nside))
        n_downsample_pix = 4**(k_downsample - k_tile)

        n_skies_per_pix = (downsample // n_downsample_pix) + 1

        # Get the sky positions per downsampled tile.
        valid_skies_downsampled = (valid_skies.groupby('down_pix', axis=0)
                                   .apply(lambda x: x.sample(n=(n_skies_per_pix
                                                                if len(x) > n_skies_per_pix
                                                                else len(x)),
                                                             random_state=seed))
                                   .reset_index(drop=True))

    elif isinstance(downsample, (list, tuple, numpy.ndarray)):

        valid_skies_downsampled = valid_skies.loc[valid_skies[cpcol].isin(downsample), :]

    else:

        raise ValueError('invalid format for downsample.')

    if len(valid_skies_downsampled) > 0:
        valid_skies_downsampled.set_index(cpcol, inplace=True)
    else:
        return False

    return valid_skies_downsampled


def get_sky_table(database, table, output, tiles=None, append=False,
                  tile_nside=32, candidate_nside=32768, min_separation=10,
                  chunk_tiles=5, ra_column='ra', dec_column='dec',
                  mag_column=None, is_flux=False, mag_threshold=None,
                  downsample=False, downsample_nside=256, n_cpus=1, seed=None):
    r"""Identifies skies from a database table.

    Skies are selected using the following procedure:

    - The sky is divided in HEALPix "tiles" of nside ``tile_nside``. For each
      tile the catalogue is queried to retrieve all the targets that lie in
      the tile footprint. This assumes that the ``healpix_ang2ipix_nest``
      Postgresql function from `pg_healpix
      <https://github.com/segasai/pg_healpix>`__ is available. It's recommended
      that an index is created for the tiling norder to speed the query.

    - Each tile is subsequently divided in pixels of nside ``candidate_nside``.
      Pixels that contain a catalogue target are removed as possible sky
      candidates. For the remaining pixels a nearest-neighbour search is done
      using a KDTree algorithm. Those pixels with neighbours closer than
      ``min_separation`` are rejected. The remaining pixels are considered
      valid skies. Alternatively, if ``mag_column`` and ``mag_threshold`` are
      defined, the separations to neighbours brighter than the threshold
      magnitude is corrected using the expression
      :math:`s^* = s + \sqrt{\dfrac{m_{thr}-m}{a}}` where
      :math:`s` is the minimum separation, :math:`m_{thr}` is the magnitude
      threshold, :math:`a=0.2` is a factor that takes into account the average
      seeing for APO and LCO.

    - If ``downsample`` is an integer, only that number of skies are returned
      for each tile. If so, the tile is divided in pixels of nside
      ``downsample_nside`` (which should be smaller than ``candidate_nside``
      but larger than ``tile_nside``) and for each downsample pixel
      ``int(downsample / downsample_npix) + 1`` skies are selected (or all the
      skies, if fewer are available), where ``downsample_npix`` is the number
      of pixels corresponding to the ``downsample_nside`` nside.

    - The process can be parallelised for each tile and the results are
      compiled in a single Pandas data frame that is saved to an HDF5 file.

    All pixel values use the nested ordering.

    Parameters
    ----------
    database : ~sdssdb.connection.PeeweeDatabaseConnection
        A valid database connection.
    table : str
        Name of the table to query, optionally schema-qualified.
    output : str
        Path to the HDF5 file to write.
    tiles : list
        A list of HEALPix pixels of nside ``tile_nside`` for which the sky
        selection will be done. If `None`, runs for all the pixels of nside
        ``tile_nside``.
    append : bool
        If `True`, appends to the output file if it exists. Otherwise
        overwrites it.
    tile_nside : int
        The HEALPix nside to use to tile the all-sky catalogue.
    candidate_nside : int
        The HEALPix nside used to identify candidate pixels in each tile.
        Candidates are then checked to confirm that their closest neighbour
        is at least ``min_separation`` arcsec away.
    min_separation : int
        The minimum separation, in arcsec, between skies and their closest
        neighbour in the catalogue.
    chunk_tiles : int
        Request the targets from this number of tiles at once, and then process
        them one by one. Reduces the number of queries at the expense of making
        each one return more rows and use more memory. This option is ignored
        if ``tiles`` is set.
    ra_column : str
        The name of the column in ``table`` that contains the Right Ascension
        coordinates, in degrees.
    dec_column : str
        The name of the column in ``table`` that contains the Declination
        coordinates, in degrees.
    mag_column : str
        The name of the column in ``table`` with the magnitude to be used to
        scale ``min_separation``.
    is_flux : bool
        If `True`, assumes the ``mag_column`` values are given in nanomaggies.
    mag_threshold : float
        The value below which the separation to neighbouring sources will be
        scaled.
    downsample : bool, int, or list
        The total number of skies to retrieve for each tile. If `False`,
        returns all candidate skies. ``downsample`` can also be a list of
        pixels with nside ``candidate_nside`` in which case only pixels in
        that list will be returned.
    downsample_nside : int
        The HEALPix nside used for downsampling. If ``downsample=True``, the
        resulting valid skies will be grouped by HEALPix pixels of this
        resolution. For each pixel a random sample will be drawn so that the
        total number of skies selected matches ``downsample``.
    n_cpus : int
        Number of CPUs to use for multiprocessing.
    seed : int
        The random state seed.

    """

    assert database.connected, 'database is not connected.'

    columns = (f'healpix_ang2ipix_nest('
               f'{tile_nside}, {ra_column}, {dec_column}) AS tile_{tile_nside}, '
               f'{ra_column}, {dec_column}')
    if mag_column and mag_threshold:
        columns += f', {mag_column}'

    if chunk_tiles is None:
        chunk_tiles = 1

    if tiles is None:
        tiles = numpy.arange(healpy.nside2npix(tile_nside))

    query = (f'SELECT {columns} FROM {table} '
             f'WHERE healpix_ang2ipix_nest('
             f'{tile_nside}, {ra_column}, {dec_column}) IN ({{values}}) '
             f'ORDER BY tile_{tile_nside};')

    pbar = manager.counter(total=len(tiles), desc='Tiles', unit='tiles')
    pbar.refresh()

    process_tile = partial(_process_tile,
                           candidate_nside=candidate_nside,
                           tile_nside=tile_nside,
                           min_separation=min_separation,
                           mag_column=mag_column, is_flux=is_flux,
                           mag_threshold=mag_threshold,
                           downsample=downsample,
                           downsample_nside=downsample_nside,
                           seed=seed)

    all_skies = None

    for tile_split in numpy.array_split(tiles, len(tiles) / chunk_tiles):

        values = ','.join(map(str, tile_split))

        targets = pandas.read_sql(query.format(values=values), database)
        targets.rename(columns={ra_column: 'ra', dec_column: 'dec'}, inplace=True)

        if n_cpus > 1:

            groups = targets.groupby(f'tile_{tile_nside}')

            # Increment the counter for each tile that didn't get any target.
            n_groups = len(groups)
            non_hit = len(tile_split) - n_groups
            pbar.update(non_hit)

            with multiprocessing.Pool(n_cpus) as pool:

                for valid_skies in pool.imap_unordered(process_tile, groups):
                    if valid_skies is not False and len(valid_skies) > 0:
                        if all_skies is None:
                            all_skies = valid_skies
                        else:
                            all_skies = all_skies.append(valid_skies)

                    pbar.update()

        else:

            for tile in tile_split:
                valid_skies = process_tile((tile, targets))
                if valid_skies is not False:
                    if all_skies is None and len(valid_skies) > 0:
                        all_skies = valid_skies
                    else:
                        all_skies = all_skies.append(valid_skies)

                pbar.update()

    key = 'data'
    hdf = pandas.HDFStore(output, mode='w')

    if append is False and key in hdf:
        hdf.remove(key)

    hdf.append(key, all_skies)

    hdf.close()


def plot_sky_density(file_or_data, nside, nside_plot=32, **kwargs):
    """Plots the number of skies as a HEALPix map.

    Parameters
    ----------
    file_or_data : str or ~pandas.DataFrame
        A HDF5 file with the sky catalogue or a Pandas data frame with the
        data. It is assumed that the file or data frame contains at least a
        column ``pix_X``, where ``X`` is the order corresponding to ``nside``.
    nside : int
        The HEALPix nside for the skies.
    nside_plot : int
        The HEALPix nside in which the data will be plotted.
    kwargs : dict
        Other keyword parameters to pass to healpy's
        `~healpy.visufunc.mollview`.

    Returns
    -------
    figure : `~matplotlib.figure.Figure`
        The Matplotlib figure with the plot.

    """

    if isinstance(file_or_data, str):
        data = pandas.read_hdf(file_or_data)
    else:
        data = file_or_data

    k = int(numpy.log2(nside))
    k_plot = int(numpy.log2(nside_plot))

    pix_plot = f'pix_{k_plot}'
    pix = f'pix_{k}'

    data[pix_plot] = nested_regrade(data[pix], nside, nside_plot)
    count = data.groupby(pix_plot).count()

    hmap = numpy.arange(healpy.nside2npix(nside_plot), dtype=numpy.float32)
    hmap[:] = healpy.UNSEEN
    hmap[count.index] = count[pix]

    figure = plt.figure()
    healpy.mollview(hmap, fig=figure.number, nest=True, **kwargs)

    return figure


def plot_skies(file_or_data, ra, dec, radius=1.5, targets=None,
               show_sky_buffer=False, buffer_radius=10.):
    """Plots the skies (and optionally targets) in a regions.

    Parameters
    ----------
    file_or_data : str or ~pandas.DataFrame
        A HDF5 file with the sky catalogue or a Pandas data frame with the
        data, including columns labelled ``ra`` and ``dec``.
    ra : float
        The right ascension of the centre of the field.
    dec : float
        The declination of the centre of the field.
    radius : float
        The FOV radius, in degrees.
    targets : list
        A target list or array as ``(ra, dec)``.
    show_sky_buffer : bool
        Plots the buffer regions around each sky.
    buffer_radius : float
        The sky buffer radius, in arcsec.

    Returns
    -------
    figure : `~matplotlib.figure.Figure`
        The Matplotlib figure with the plot.

    """

    if isinstance(file_or_data, str):
        data = pandas.read_hdf(file_or_data)
    else:
        data = file_or_data

    fig, ax = plt.subplots()

    cos_factor = numpy.cos(numpy.radians(dec))

    FOV = Ellipse((ra, dec), radius * 2 / cos_factor,
                  radius * 2, linewidth=2, color='None',
                  ec='k')
    ax.add_patch(FOV)

    # Plot skies
    centre = SkyCoord(ra=ra, dec=dec, unit='deg')
    sky_coords = SkyCoord(ra=data.ra.to_numpy(),
                          dec=data.dec.to_numpy(),
                          unit='deg')
    skies = data[sky_coords.separation(centre).value < radius]

    ax.scatter(skies.ra, skies.dec,
               marker='o', s=0.2, color='b', zorder=10, label='Skies')

    if show_sky_buffer:
        for _, sky in skies.iterrows():
            buffer = Ellipse((sky.ra, sky.dec),
                             buffer_radius / 3600. * 2,
                             buffer_radius / 3600. * 2,
                             color='y', ec='None',
                             alpha=0.2, zorder=0)
            ax.add_patch(buffer)

    # Plot targets
    if targets is not None:
        targets = numpy.array(targets)
        target_coords = SkyCoord(ra=targets[:, 0],
                                 dec=targets[:, 1],
                                 unit='deg')
        targets = targets[target_coords.separation(centre).value < radius]

        ax.scatter(targets[:, 0], targets[:, 1], marker='x', s=0.2,
                   color='r', zorder=20, label='Targets')

    ax.legend(loc='upper right')

    ax.set_xlim(ra - (radius - 0.1) / cos_factor,
                ra + (radius + 0.1) / cos_factor)
    ax.set_ylim(dec - radius - 0.1, dec + radius + 0.1)

    ax.set_xlabel('Right Ascension [deg]')
    ax.set_ylabel('Declination [deg]')

    ax.set_aspect(1 / cos_factor)

    fig.tight_layout()

    return fig


def create_sky_catalogue(database, tiles=None, **kwargs):
    """A script to generate a combined sky catalogue from multiple sources."""

    if not os.path.exists('gaia_skies.h5'):
        log.info('Procesing gaia_dr2_source.')
        get_sky_table(database, 'catalogdb.gaia_dr2_source', 'gaia_skies.h5',
                      mag_column='phot_g_mean_mag', mag_threshold=12,
                      downsample=2000, tiles=tiles, **kwargs)
    else:
        warnings.warn('Found file gaia_skies.h5', TargetSelectionUserWarning)

    # We use Gaia as the source for the downsampled candidates.
    gaia = pandas.read_hdf('gaia_skies.h5')
    downsample_list = gaia.index.drop_duplicates().values

    if not os.path.exists('ls8_skies.h5'):
        log.info('Procesing legacy_survey_dr8.')
        get_sky_table(database, 'catalogdb.legacy_survey_dr8', 'ls8_skies.h5',
                      mag_column='flux_g', is_flux=True, mag_threshold=12,
                      downsample=downsample_list, tiles=tiles, **kwargs)
    else:
        warnings.warn('Found file ls8_skies.h5', TargetSelectionUserWarning)

    if not os.path.exists('tmass_skies.h5'):
        log.info('Procesing twomass_psc.')
        get_sky_table(database, 'catalogdb.twomass_psc', 'tmass_skies.h5',
                      dec_column='decl', mag_column='h_m', mag_threshold=12,
                      downsample=downsample_list, tiles=tiles, **kwargs)
    else:
        warnings.warn('Found file tmass_skies.h5', TargetSelectionUserWarning)

    if not os.path.exists('tycho2_skies.h5'):
        log.info('Procesing tycho2.')
        get_sky_table(database, 'catalogdb.tycho2', 'tycho2_skies.h5',
                      ra_column='ramdeg', dec_column='demdeg',
                      mag_column='vtmag', mag_threshold=12,
                      downsample=downsample_list, tiles=tiles, **kwargs)
    else:
        warnings.warn('Found file tycho2_skies.h5', TargetSelectionUserWarning)

    if not os.path.exists('tmass_xsc_skies.h5'):
        log.info('Procesing twomass_xsc.')
        get_sky_table(database, 'catalogdb.twomass_xsc', 'tmass_xsc_skies.h5',
                      dec_column='decl', mag_column='h_m_k20fe', mag_threshold=14,
                      downsample=downsample_list, tiles=tiles, **kwargs)
    else:
        warnings.warn('Found file tmass_xsc_skies.h5', TargetSelectionUserWarning)

    skies = None
    col_order = []

    for file_ in ['gaia_skies.h5', 'ls8_skies.h5', 'tmass_skies.h5',
                  'tycho2_skies.h5', 'tmass_xsc_skies.h5']:

        table_name = file_[0:-9]

        table = pandas.read_hdf(file_).drop_duplicates()
        table.rename(columns={'sep_neighbour': f'sep_neighbour_{table_name}',
                              'mag_neighbour': f'mag_neighbour_{table_name}'},
                     inplace=True)

        table.loc[:, f'{table_name}_sky'] = True

        if skies is None:
            skies = table
        else:
            skies = skies.combine_first(table)

        skies.fillna({f'{table_name}_sky': False}, inplace=True)

        col_order += [f'{table_name}_sky',
                      f'sep_neighbour_{table_name}',
                      f'mag_neighbour_{table_name}']

    skies = skies.loc[:, ['ra', 'dec', 'down_pix', 'tile_32'] + col_order]
    skies.to_hdf('skies.h5', 'data')



# encoding: utf-8

import os

import enlighten

from sdsstools import get_config, get_logger, get_package_version


NAME = 'sdss-target-selection'

config = get_config('target_selection',
                    os.path.join(os.path.dirname(__file__),
                                 'config/target_selection.yml'))

log = get_logger(NAME)


__version__ = get_package_version(path=__file__,
                                  package_name=NAME,
                                  pep_440=True)


manager = enlighten.get_manager()


from .xmatch import XMatchPlanner, XMatchModel  # isort:skip



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Pramod Gupta (psgupta@uw.edu)
# @Date: 2020-11-14
# @Filename: ops_apogee_stds.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee

from sdssdb.peewee.sdss5db.catalogdb import (Catalog, CatalogToTIC_v8,
                                             TIC_v8, TwoMassPSC)

from target_selection.cartons import BaseCarton


# See catalog.py for the name of peewee model names corresponding
# to postgres table names:
# https://github.com/sdss/sdssdb/blob/master/python/sdssdb/peewee/sdss5db/catalogdb.py
#
# peewee Model name ---> postgres table name
# TwoMassPSC(CatalogdbModel) --->'twomass_psc'
#
# The CatalogTo* peewee model names are not explicit in catalogdb.py.
# The names are constructed by prepending CatalogTo
# to the corresponding model name which is in catalogdb.py.
# For example:
# CatalogToTIC_v8--->'catalog_to_tic_v8'

class OPS_APOGEE_Stds_Carton(BaseCarton):
    """
    Shorthand name: ops_apogee_stds
    Selection criteria:
    Shorthand name: ops_apogee_stds

    Simplified Description of selection criteria:
    parent catalog is 2MASS point source catalog, after restricting to:

    7 < H < 11,
    -0.25 < J - K < 0.3,
    j_msigcom < 0.1, h_msigcom < 0.1, k_msigcom < 0.1,
    rd_flg = 2 or 1 in all three (JHK) bands
    ph_qual = A or B in all three (JHK) bands
    gal_contam = 0 (no galaxy contamination)
    prox > 6.0
    cc_flg = 000 (zero in all three (JHK) bands)
    ext_key = null (no match in extended source catalog)
    Once 2MASS has been limited to sources meeting the above criteria,
    the sky is then divided into an nside = 128 HEALPIX skymap.
    The 5 bluest sources (i.e. 5 smallest J-K) in each healpix are then
    selected and saved for the output carton.

    (note – the HEALPIX binning and sub-querying might be useful
    to implement as a function that can be run for future cartons;
    I can imagine that this capability might prove useful
    for updated BOSS calibrator cartons, and
    maybe for searching for guide stars, or
    even maybe some kind of science cartons that
    depend on sampling things smoothly in a spatial sense?)

    All HEALPix use the nested ordering.
    number_healpix_pixels = 12 *Nside*Nside
    Hence number_healpix_pixels = 12*128*128 = 196608

    Return columns: All the filter columns plus healpix_128
    Lead contact: Kevin Covey
    """

    name = 'ops_std_apogee'
    category = 'standard'
    cadence = None
    program = 'ops_std'
    priority = 5500
    mapper = None

    def build_query(self, version_id, query_region=None):

        query = (Catalog
                 .select(CatalogToTIC_v8.catalogid, Catalog.ra, Catalog.dec,
                         TwoMassPSC.h_m, TwoMassPSC.j_m, TwoMassPSC.k_m,
                         TwoMassPSC.h_msigcom,
                         TwoMassPSC.j_msigcom,
                         TwoMassPSC.k_msigcom,
                         TwoMassPSC.rd_flg, TwoMassPSC.ph_qual,
                         TwoMassPSC.gal_contam, TwoMassPSC.prox,
                         TwoMassPSC.cc_flg, TwoMassPSC.ext_key,
                         peewee.fn.healpix_ang2ipix_nest(
                             128, Catalog.ra, Catalog.dec).alias('healpix_128'),
                         (TwoMassPSC.j_m - TwoMassPSC.k_m).alias('j_minus_k'))
                 .join(CatalogToTIC_v8,
                       on=(Catalog.catalogid == CatalogToTIC_v8.catalogid))
                 .join(TIC_v8,
                       on=(CatalogToTIC_v8.target_id == TIC_v8.id))
                 .join(TwoMassPSC,
                       on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True,
                        TwoMassPSC.h_m > 7,
                        TwoMassPSC.h_m < 11,
                        (TwoMassPSC.j_m - TwoMassPSC.k_m) > -0.25,
                        (TwoMassPSC.j_m - TwoMassPSC.k_m) < 0.5,
                        TwoMassPSC.j_msigcom < 0.1,
                        TwoMassPSC.h_msigcom < 0.1,
                        TwoMassPSC.k_msigcom < 0.1,
                        (TwoMassPSC.rd_flg == '111') |
                        (TwoMassPSC.rd_flg == '211') |
                        (TwoMassPSC.rd_flg == '121') |
                        (TwoMassPSC.rd_flg == '221') |
                        (TwoMassPSC.rd_flg == '112') |
                        (TwoMassPSC.rd_flg == '212') |
                        (TwoMassPSC.rd_flg == '122') |
                        (TwoMassPSC.rd_flg == '222'),
                        (TwoMassPSC.ph_qual == 'AAA') |
                        (TwoMassPSC.ph_qual == 'BAA') |
                        (TwoMassPSC.ph_qual == 'ABA') |
                        (TwoMassPSC.ph_qual == 'BBA') |
                        (TwoMassPSC.ph_qual == 'AAB') |
                        (TwoMassPSC.ph_qual == 'BAB') |
                        (TwoMassPSC.ph_qual == 'ABB') |
                        (TwoMassPSC.ph_qual == 'BBB'),
                        TwoMassPSC.gal_contam == 0,
                        TwoMassPSC.prox > 6.0,
                        TwoMassPSC.cc_flg == '000',
                        TwoMassPSC.ext_key.is_null()))

        # Below ra, dec and radius are in degrees
        # query_region[0] is ra of center of the region
        # query_region[1] is dec of center of the region
        # query_region[2] is radius of the region
        if query_region:
            query = (query
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))
        return query

    def post_process(self, model):
        """
        Select the 5 bluest sources (i.e. 5 smallest J-K) in each healpix pixel.
        """

        self.database.execute_sql("update sandbox.temp_ops_std_apogee " +
                                  "set selected = false")

        cursor = self.database.execute_sql(
            "select catalogid, healpix_128, j_minus_k from " +
            " sandbox.temp_ops_std_apogee " +
            " order by healpix_128 asc, j_minus_k asc;")

        output = cursor.fetchall()

        list_of_catalog_id = [0] * len(output)
        nside = 128
        total_number_healpix_pixels = 12 * nside * nside
        count = [0] * total_number_healpix_pixels
        current_target = 0
        for i in range(len(output)):
            current_healpix = output[i][1]
            if(count[current_healpix] < 5):
                count[current_healpix] = count[current_healpix] + 1
                list_of_catalog_id[current_target] = output[i][0]
                current_target = current_target + 1

        max_target = current_target
        for k in range(max_target + 1):
            self.database.execute_sql(
                " update sandbox.temp_ops_std_apogee set selected = true " +
                " where catalogid = " + str(list_of_catalog_id[k]) + ";")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-05-29
# @Filename: mwm_galactic.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee

from sdssdb.peewee.sdss5db.catalogdb import (Catalog, CatalogToTIC_v8,
                                             TIC_v8, TwoMassPSC)

from . import BaseCarton


class MWM_Galactic_Genesis_Carton(BaseCarton):
    """Galactic Genesis carton.

    Definition: Selection of all IR-bright, red stars – vast majority are red
    giants; follows the density distribution of the MW (concentrated in the
    plane and bulge). Select sources brighter than H<11 AND ((G-H) > 3.5 OR
    Gaia non-detection). Approximately 5 million stars.

    Pseudo-query:

        SELECT catalog.catalogid, catalog.ra, catalog.ra, catalog.dec,
            catalog.pmra, catalog.pmdec, catalog.epoch, twomass_psc.h_m,
            gaia_dr2_source.phot_g_mean_mag
        FROM catalog
        JOIN catalog_to_TIC, TIC, gaia_dr2_source, gaia_clean, twomass_psc
        WHERE h_m < 11
            AND (ph_qual[1] = 'A' OR ph_qual[1] = 'B')
            AND gal_contam=0
            AND cc_flg[1]=0
            AND (rd_flg[1] > 0 AND rd_flg[1] <= 3)
            AND [(phot_g_mean_mag-h_m) > 3.5  OR NOT_EXISTS(phot_g_mean_mag)]

    """

    name = 'mwm_gg_core'
    category = 'science'
    cadence = 'mwm_galactic_1x1'
    priority = 2710
    program = 'mwm_gg'
    mapper = 'MWM'

    def build_query(self, version_id, query_region=None):

        Hmag = TIC_v8.hmag
        Gmag = TIC_v8.gaiamag

        ph_qual = TwoMassPSC.ph_qual
        cc_flg = TwoMassPSC.cc_flg
        rd_flg = TwoMassPSC.rd_flg
        rd_flag_1 = peewee.fn.substr(rd_flg, 2, 1).cast('integer')

        gal_contam = TwoMassPSC.gal_contam

        h_max = self.parameters['h_max']
        g_h = self.parameters['g_h']

        gg = (TIC_v8
              .select(CatalogToTIC_v8.catalogid,
                      Hmag, Gmag,
                      ph_qual, cc_flg, rd_flg, gal_contam)
              .join(TwoMassPSC)
              .join_from(TIC_v8, CatalogToTIC_v8)
              .where(Hmag < h_max,
                     ph_qual.regexp('.(A|B).'),
                     gal_contam == 0,
                     peewee.fn.substr(cc_flg, 2, 1) == '0',
                     rd_flag_1 > 0, rd_flag_1 <= 3,
                     ((Gmag - Hmag) > g_h) | (TIC_v8.gaia >> None))
              .where(CatalogToTIC_v8.version_id == version_id,
                     CatalogToTIC_v8.best >> True))

        if query_region:
            gg = (gg
                  .join_from(CatalogToTIC_v8, Catalog)
                  .where(peewee.fn.q3c_radial_query(Catalog.ra, Catalog.dec,
                                                    query_region[0],
                                                    query_region[1],
                                                    query_region[2])))

        return gg



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Tom Dwelly
# @Date: 2020-03-24
# @Filename: masking.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

# flake8: noqa
# isort: skip_file

import os
import numpy as np
import healpy as hp
from pymoc import MOC
import pymangle as mangle
from astropy.io import fits
from astropy.coordinates import SkyCoord, search_around_sky
from astropy import units as u


known_masktypes = ["mangle", "hpxmoc", "circles", ]
known_senses = ["include", "exclude", ]




class SkyMask(object):
    '''
    This is a generic masking class that carries out selections of targets based on their RA,DEC

    Parameters
    ==========

    filename     - filename (no path) of the mask description
                 - will look only inside the target_selection/python/target_selection/masks/ directory

    masktype     - variety of mask description
                 - must be in ["mangle", "hpxmoc", "circles", ]

    sense        - "include" or "exclude" objects inside this mask?

    name         - Nickname for information only

    col_lon     - used by "circles" masks - gives the column name containing the longitude information
    col_lat      - used by "circles" masks - gives the column name containing the latitude information
    radius       - used by "circles" masks (assumed to be in degrees)
                   - if a string then gives the column name containing the radius information
                   - if a float then assume a uniform radius for all circles

    '''

    def __init__(self,
                 filename=None,
                 masktype="mangle",
                 sense="include",
                 name=None,
                 col_lon="racen",
                 col_lat="deccen",
                 radius="radius"):
        assert filename is not None
        assert len(filename) > 0
        self.filename = os.path.expanduser(filename)

        self.masktype = masktype.lower()
        assert self.masktype in known_masktypes

        self.sense = sense.lower()
        assert self.sense in known_senses

        self.col_lon = col_lon
        self.col_lat = col_lat
        self.radius  = radius

        if name is not None:
            self.name = name
        else:
            self.name = self.filename

    def __str__(self):
        return f"SkyMask({self.masktype}) {self.name}"

    def apply_mangle(self, lon=None, lat=None):

        try:
            m = mangle.Mangle(self.filename)
        except:
            raise Exception(f"Unable to open mangle mask file: {self.filename}")

        m_mask = m.contains(lon, lat)

        return m_mask


    def apply_hpxmoc(self, lon=None, lat=None):

        try:
            moc = MOC()
            moc.read(self.filename, filetype="fits")
        except:
            raise Exception(f"Unable to find/open Healpix MOC file: {self.filename}")

        # get the moc nside at the max resolution of the MOC
        nside = hp.order2nside(moc.order)
        #get the healpix pixels indices of the targets at the max resolution
        idx = hp.ang2pix(nside, lon, lat, lonlat=True, nest=True )

        m_mask = moc.contains(idx)
        return m_mask


    def apply_circles(self, lon=None, lat=None):

        # read the list of circles from the file

        try:
            hdul = fits.open(self.filename)
        except:
            raise Exception(f"Failed to find/open circles file: {self.filename}")

        data = hdul[1].data

        try:
            circle_lon = data[self.col_lon]
            circle_lat = data[self.col_lat]
        except:
            raise Exception(f"Failed to find circle columns: {self.filename}[1][{self.col_lon},{self.col_lat}]")

        if isinstance(self.radius, float):
            circle_radius = np.full(len(data), self.radius)
        elif isinstance(self.radius, str):
            try:
                circle_radius = data[self.radius]
            except:
                raise Exception(f"Failed to find circle column: {self.filename}[1][{self.radius}]")
        else:
            raise Exception(f"Cannot interpret radius: {self.radius}")

        hdul.close()

        ### need to use a single radius for all objects
        seplimit = u.Quantity(circle_radius.max(), unit='deg')

        coords_c = SkyCoord(circle_lon, circle_lat, frame="icrs", unit="deg")
        coords_t = SkyCoord(lon, lat, frame="icrs", unit="deg")
        idx_t, idx_c, d2d, d3d  = search_around_sky(coords_t, coords_c, seplimit)


        ## filter the list of matches to use the per-circle radius
        i_t_ok = [ i_t for i_t,i_c,d in zip(idx_t, idx_c, d2d) if d.deg < circle_radius[i_c]]

        m_mask = np.zeros(len(lon), np.bool)
        m_mask[i_t_ok] = True

        return m_mask




    def apply(self, lon=None, lat=None, flags=None):

        nlon = len(lon) if hasattr(lon, "__len__") else 1
        nlat = len(lat) if hasattr(lat, "__len__") else 1

        assert nlon == nlat
        if nlon == 0:
            return None

        if flags is not None:
            nflags = len(flags) if hasattr(flags, "__len__") else 1
            assert nflags == nlat

        if self.masktype == "mangle":
            m_mask = self.apply_mangle(lon, lat)
        elif self.masktype == "hpxmoc":
            m_mask = self.apply_hpxmoc(lon, lat)
        elif self.masktype == "circles":
            m_mask = self.apply_circles(lon, lat)
        else:
            raise Exception(f"Unknown masktype: {self.masktype}")

        # if necessary, adjust the sense of the sky mask
        if self.sense == "exclude":
            m_mask = ~m_mask

        # if input flags were supplied then do a logical AND with them
        if flags is None:
            return m_mask
        else:
            return m_mask & flags
